# coding=utf-8
#
# Copyright (C) 2006 Jean-Francois Barraud, barraud@math.univ-lille1.fr
# Copyright (C) 2010 Alvin Penner, penner@vaxxine.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# barraud@math.univ-lille1.fr
#
# This code defines several functions to make handling of transform
# attribute easier.
#
"""
Provide transformation parsing to extensions
"""
from __future__ import annotations
import re
from decimal import Decimal
from math import cos, radians, sin, sqrt, tan, fabs, atan2, hypot, pi, isfinite
from typing import (
    overload,
    cast,
    Callable,
    Generator,
    Iterator,
    Tuple,
    Union,
    Optional,
    List,
)


from .utils import strargs, KeyDict


VectorLike = Union[
    "ImmutableVector2d", Tuple[float, float]
]  # pylint: disable=invalid-name
MatrixLike = Union[
    str,
    Tuple[Tuple[float, float, float], Tuple[float, float, float]],
    Tuple[float, float, float, float, float, float],
    "Transform",
]
BoundingIntervalArgs = Union[
    "BoundingInterval", Tuple[float, float], float
]  # pylint: disable=invalid-name

# All the names that get added to the inkex API itself.
__all__ = (
    "BoundingBox",
    "DirectedLineSegment",
    "ImmutableVector2d",
    "Transform",
    "Vector2d",
)


# Old settings, supported because users click 'ok' without looking.
XAN = KeyDict({"l": "left", "r": "right", "m": "center_x"})
YAN = KeyDict({"t": "top", "b": "bottom", "m": "center_y"})
# Anchoring objects with given directions (see inx options)
CUSTOM_DIRECTION = {270: "tb", 90: "bt", 0: "lr", 360: "lr", 180: "rl"}
DIRECTION = ["tb", "bt", "lr", "rl", "ro", "ri"]


class ImmutableVector2d:
    """Represents an immutable element of 2-dimensional Euclidean space"""

    _x = 0.0
    _y = 0.0

    x = property(lambda self: self._x)
    y = property(lambda self: self._y)

    @overload
    def __init__(self):
        # type: () -> None
        pass

    @overload
    def __init__(self, v, fallback=None):
        # type: (Union[VectorLike, str], Optional[Union[VectorLike, str]]) -> None
        pass

    @overload
    def __init__(self, x, y):
        # type: (float, float) -> None
        pass

    def __init__(self, *args, fallback=None):
        try:
            if len(args) == 0:
                x, y = 0.0, 0.0
            elif len(args) == 1:
                x, y = self._parse(args[0])
            elif len(args) == 2:
                x, y = map(float, args)
            else:
                raise ValueError("too many arguments")
        except (ValueError, TypeError) as error:
            if fallback is None:
                raise ValueError("Cannot parse vector and no fallback given") from error
            x, y = ImmutableVector2d(fallback)
        self._x, self._y = float(x), float(y)

    @staticmethod
    def _parse(point):
        # type: (Union[VectorLike, str]) -> Tuple[float, float]
        if isinstance(point, ImmutableVector2d):
            x, y = point.x, point.y
        elif isinstance(point, (tuple, list)) and len(point) == 2:
            x, y = map(float, point)
        elif isinstance(point, str) and point.count(",") == 1:
            x, y = map(float, point.split(","))
        else:
            raise ValueError(f"Can't parse {repr(point)}")
        return x, y

    def __add__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        return Vector2d(self.x + other.x, self.y + other.y)

    def __radd__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        return Vector2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        return Vector2d(self.x - other.x, self.y - other.y)

    def __rsub__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        return Vector2d(-self.x + other.x, -self.y + other.y)

    def __neg__(self):
        # type: () -> Vector2d
        return Vector2d(-self.x, -self.y)

    def __pos__(self):
        # type: () -> Vector2d
        return Vector2d(self.x, self.y)

    def __floordiv__(self, factor):
        # type: (float) -> Vector2d
        return Vector2d(self.x / float(factor), self.y / float(factor))

    def __truediv__(self, factor):
        # type: (float) -> Vector2d
        return Vector2d(self.x / float(factor), self.y / float(factor))

    def __div__(self, factor):
        # type: (float) -> Vector2d
        return Vector2d(self.x / float(factor), self.y / float(factor))

    def __mul__(self, factor):
        # type: (float) -> Vector2d
        return Vector2d(self.x * factor, self.y * factor)

    def __abs__(self):
        # type: () -> float
        return self.length

    def __rmul__(self, factor):
        # type: (float) -> VectorLike
        return Vector2d(self.x * factor, self.y * factor)

    def __repr__(self):
        # type: () -> str
        return f"Vector2d({self.x:.6g}, {self.y:.6g})"

    def __str__(self):
        # type: () -> str
        return f"{self.x:.6g}, {self.y:.6g}"

    def __iter__(self) -> Generator[float, None, None]:
        yield self.x
        yield self.y

    def __len__(self):
        # type: () -> int
        return 2

    def __getitem__(self, item):
        # type: (int) -> float
        return (self.x, self.y)[item]

    def to_tuple(self) -> Tuple[float, float]:
        """A tuple of the vector's components"""
        return cast(Tuple[float, float], tuple(self))

    def to_polar_tuple(self):
        # type: () -> Tuple[float, Optional[float]]
        """A tuple of the vector's magnitude and direction

        .. versionadded:: 1.1"""
        return self.length, self.angle

    def dot(self, other: VectorLike) -> float:
        """Multiply Vectors component-wise"""
        other = Vector2d(other)
        return self.x * other.x + self.y * other.y

    def cross(self, other):
        # type: (VectorLike) -> float
        """Z component of the cross product of the vectors extended into 3D

        .. versionadded:: 1.1"""
        other = Vector2d(other)
        return self.x * other.y - self.y * other.x

    def is_close(
        self,
        other: Union[VectorLike, str, Tuple[float, float]],
        rtol: float = 1e-5,
        atol: float = 1e-8,
    ) -> float:
        """Checks if two vectors are (almost) identical, up to both absolute and
        relative tolerance."""
        other = Vector2d(other)
        delta = (self - other).length
        return delta < (atol + rtol * other.length)

    @property
    def length(self) -> float:
        """Returns the length of the vector"""
        return sqrt(self.dot(self))

    @property
    def angle(self):
        # type: () -> Optional[float]
        """The angle of the vector when represented in polar coordinates

        .. versionadded:: 1.1"""
        if self.x == 0 and self.y == 0:
            return None
        return atan2(self.y, self.x)


class Vector2d(ImmutableVector2d):
    """Represents an element of 2-dimensional Euclidean space"""

    @staticmethod
    def from_polar(radius, theta):
        # type: (float, Optional[float]) -> Optional[Vector2d]
        """Creates a Vector2d from polar coordinates

        None is returned when theta is None and radius is not zero.

        .. versionadded:: 1.1
        """
        if radius == 0.0:
            return Vector2d(0.0, 0.0)
        if theta is not None:
            return Vector2d(radius * cos(theta), radius * sin(theta))
        # A vector with a radius but no direction is invalid
        return None

    @ImmutableVector2d.x.setter
    def x(self, value):
        # type: (Union[float, int, str]) -> None
        self._x = float(value)

    @ImmutableVector2d.y.setter
    def y(self, value):
        # type: (Union[float, int, str]) -> None
        self._y = float(value)

    def __iadd__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        self.x += other.x
        self.y += other.y
        return self

    def __isub__(self, other):
        # type: (VectorLike) -> Vector2d
        other = Vector2d(other)
        self.x -= other.x
        self.y -= other.y
        return self

    def __imul__(self, factor):
        # type: (float) -> Vector2d
        self.x *= factor
        self.y *= factor
        return self

    def __idiv__(self, factor):
        # type: (float) -> Vector2d
        self.x /= factor
        self.y /= factor
        return self

    def __itruediv__(self, factor):
        # type: (float) -> Vector2d
        self.x /= factor
        self.y /= factor
        return self

    def __ifloordiv__(self, factor):
        # type: (float) -> Vector2d
        self.x /= factor
        self.y /= factor
        return self

    @overload
    def assign(self, x, y):
        # type: (float, float) -> VectorLike
        pass

    @overload
    def assign(self, other):
        # type: (VectorLike, str) -> VectorLike
        pass

    def assign(self, *args):
        """Assigns a different vector in place"""
        self.x, self.y = Vector2d(*args)
        return self


class Transform:
    """A transformation object which will always reduce to a matrix and can
    then be used in combination with other transformations for reducing
    finding a point and printing svg ready output.

    Use with svg transform attribute input:

      tr = Transform("scale(45, 32)")

    Use with triad matrix input (internal representation):

      tr = Transform(((1.0, 0.0, 0.0), (0.0, 1.0, 0.0)))

    Use with hexad matrix input (i.e. svg matrix(...)):

      tr = Transform((1.0, 0.0, 0.0, 1.0, 0.0, 0.0))

    Once you have a transformation you can operate tr * tr to compose,
    any of the above inputs are also valid operators for composing.
    """

    TRM = re.compile(r"(translate|scale|rotate|skewX|skewY|matrix)\s*\(([^)]*)\)\s*,?")
    absolute_tolerance = 1e-5  # type: float

    def __init__(
        self,
        matrix=None,  # type: Optional[MatrixLike]
        callback=None,  # type: Optional[Callable[[Transform], Transform]]
        **extra,
    ):
        # type: (...) -> None
        self.callback = None
        self.matrix = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0))
        if matrix is not None:
            self._set_matrix(matrix)

        self.add_kwargs(**extra)
        # Set callback last, so it doesn't kick off just setting up the internal value
        self.callback = callback

    def _set_matrix(self, matrix):
        # type: (MatrixLike) -> None
        """Parse a given string as an svg transformation instruction.

        .. versionadded:: 1.1"""
        if isinstance(matrix, str):
            for func, values in self.TRM.findall(matrix.strip()):
                getattr(self, "add_" + func.lower())(*strargs(values))
        elif isinstance(matrix, Transform):
            self.matrix = matrix.matrix
        elif isinstance(matrix, (tuple, list)) and len(matrix) == 2:
            row1 = matrix[0]
            row2 = matrix[1]
            if isinstance(row1, (tuple, list)) and isinstance(row2, (tuple, list)):
                if len(row1) == 3 and len(row2) == 3:
                    row1 = cast(Tuple[float, float, float], tuple(map(float, row1)))
                    row2 = cast(Tuple[float, float, float], tuple(map(float, row2)))
                    self.matrix = row1, row2
                else:
                    raise ValueError(
                        f"Matrix '{matrix}' is not a valid transformation matrix"
                    )
            else:
                raise ValueError(
                    f"Matrix '{matrix}' is not a valid transformation matrix"
                )
        elif isinstance(matrix, (list, tuple)) and len(matrix) == 6:
            tmatrix = cast(
                Union[List[float], Tuple[float, float, float, float, float, float]],
                matrix,
            )
            row1 = (float(tmatrix[0]), float(tmatrix[2]), float(tmatrix[4]))
            row2 = (float(tmatrix[1]), float(tmatrix[3]), float(tmatrix[5]))
            self.matrix = row1, row2
        elif not isinstance(matrix, (list, tuple)):
            raise ValueError(f"Invalid transform type: {type(matrix).__name__}")
        else:
            raise ValueError(f"Matrix '{matrix}' is not a valid transformation matrix")

    # These provide quick access to the svg matrix:
    #
    # [ a, c, e ]
    # [ b, d, f ]
    #
    a = property(lambda self: self.matrix[0][0])  # pylint: disable=invalid-name
    b = property(lambda self: self.matrix[1][0])  # pylint: disable=invalid-name
    c = property(lambda self: self.matrix[0][1])  # pylint: disable=invalid-name
    d = property(lambda self: self.matrix[1][1])  # pylint: disable=invalid-name
    e = property(lambda self: self.matrix[0][2])  # pylint: disable=invalid-name
    f = property(lambda self: self.matrix[1][2])  # pylint: disable=invalid-name

    def __bool__(self):
        # type: () -> bool
        return not self.__eq__(Transform())

    __nonzero__ = __bool__

    @overload
    def add_matrix(self, a):
        # type: (MatrixLike) -> Transform
        pass

    @overload
    def add_matrix(  # pylint: disable=too-many-arguments
        self, a: float, b: float, c: float, d: float, e: float, f: float
    ) -> Transform:
        pass

    @overload
    def add_matrix(self, a, b):
        # type: (Tuple[float, float, float], Tuple[float, float, float]) -> Transform
        pass

    def add_matrix(self, *args):
        """Add matrix in order they appear in the svg hexad"""
        if len(args) == 1:
            self.__imatmul__(Transform(args[0]))
        elif len(args) == 2 or len(args) == 6:
            self.__imatmul__(Transform(args))
        else:
            raise ValueError(f"Invalid number of arguments {args}")
        return self

    def add_kwargs(self, **kwargs):
        """Add translations, scales, rotations etc using key word arguments"""
        for key, value in reversed(list(kwargs.items())):
            func = getattr(self, "add_" + key)
            if isinstance(value, tuple):
                func(*value)
            elif value is not None:
                func(value)
        return self

    @overload
    def add_translate(self, dr):
        # type: (VectorLike) -> Transform
        pass

    @overload
    def add_translate(self, tr_x, tr_y=0.0):
        # type: (float, Optional[float]) -> Transform
        pass

    def add_translate(self, *args):
        """Add translate to this transformation"""
        if len(args) == 1 and isinstance(args[0], (int, float)):
            tr_x, tr_y = args[0], 0.0
        else:
            tr_x, tr_y = Vector2d(*args)
        self.__imatmul__(((1.0, 0.0, tr_x), (0.0, 1.0, tr_y)))
        return self

    def add_scale(self, sc_x, sc_y=None):
        """Add scale to this transformation"""
        sc_y = sc_x if sc_y is None else sc_y
        self.__imatmul__(((sc_x, 0.0, 0.0), (0.0, sc_y, 0.0)))
        return self

    @overload
    def add_rotate(self, deg, center):
        # type: (float, VectorLike) -> Transform
        pass

    @overload
    def add_rotate(self, deg, center_x, center_y):
        # type: (float, float, float) -> Transform
        pass

    @overload
    def add_rotate(self, deg):
        # type: (float) -> Transform
        pass

    @overload
    def add_rotate(self, deg, a):
        # type: (float, Union[VectorLike, str]) -> Transform
        pass

    @overload
    def add_rotate(self, deg, a, b):
        # type: (float, float, float) -> Transform
        pass

    def add_rotate(self, deg, *args):
        """Add rotation to this transformation"""
        center_x, center_y = Vector2d(*args)
        _cos, _sin = cos(radians(deg)), sin(radians(deg))
        self.__imatmul__(((_cos, -_sin, center_x), (_sin, _cos, center_y)))
        self.__imatmul__(((1.0, 0.0, -center_x), (0.0, 1.0, -center_y)))
        return self

    def add_skewx(self, deg):
        # type: (float) -> Transform
        """Add skew x to this transformation"""
        self.__imatmul__(((1.0, tan(radians(deg)), 0.0), (0.0, 1.0, 0.0)))
        return self

    def add_skewy(self, deg):
        # type: (float) -> Transform
        """Add skew y to this transformation"""
        self.__imatmul__(((1.0, 0.0, 0.0), (tan(radians(deg)), 1.0, 0.0)))
        return self

    def to_hexad(self):
        # type: () -> Iterator[float]
        """Returns the transform as a hexad matrix (used in svg)"""
        return (val for lst in zip(*self.matrix) for val in lst)

    def is_translate(self, exactly=False):
        # type: (bool) -> bool
        """Returns True if this transformation is ONLY translate"""
        tol = self.absolute_tolerance if not exactly else 0.0
        return (
            fabs(self.a - 1) <= tol
            and abs(self.d - 1) <= tol
            and fabs(self.b) <= tol
            and fabs(self.c) <= tol
        )

    def is_scale(self, exactly=False):
        # type: (bool) -> bool
        """Returns True if this transformation is ONLY scale"""
        tol = self.absolute_tolerance if not exactly else 0.0
        return (
            fabs(self.e) <= tol
            and fabs(self.f) <= tol
            and fabs(self.b) <= tol
            and fabs(self.c) <= tol
        )

    def is_rotate(self, exactly=False):
        # type: (bool) -> bool
        """Returns True if this transformation is ONLY rotate"""
        tol = self.absolute_tolerance if not exactly else 0.0
        return (
            self._is_URT(exactly=exactly)
            and fabs(self.e) <= tol
            and fabs(self.f) <= tol
            and fabs(self.a**2 + self.b**2 - 1) <= tol
        )

    def rotation_degrees(self):
        # type: () -> float
        """Return the amount of rotation in this transform"""
        if not self._is_URT(exactly=False):
            raise ValueError(
                "Rotation angle is undefined for non-uniformly scaled or skewed "
                "matrices"
            )
        return atan2(self.b, self.a) * 180 / pi

    def __str__(self):
        # type: () -> str
        """Format the given matrix into a string representation for svg"""
        hexad = tuple(self.to_hexad())
        if self.is_translate():
            if not self:
                return ""
            return f"translate({self.e:.6g}, {self.f:.6g})"
        if self.is_scale():
            return f"scale({self.a:.6g}, {self.d:.6g})"
        if self.is_rotate():
            return f"rotate({self.rotation_degrees():.6g})"
        return f"matrix({' '.join(f'{var:.6g}' for var in hexad)})"

    def __repr__(self) -> str:
        """String representation of this object"""
        return (
            f"{type(self).__name__}(("
            f"({', '.join(f'{var:.6g}' for var in self.matrix[0])}), "
            f"({', '.join(f'{var:.6g}' for var in self.matrix[1])})))"
        )

    def __eq__(self, matrix):
        # typing this requires writing a proof for mypy that matrix is really
        # MatrixLike
        """Test if this transformation is equal to the given matrix"""
        if isinstance(matrix, (str, tuple, list, Transform)):
            val = all(
                fabs(l - r) <= self.absolute_tolerance
                for l, r in zip(self.to_hexad(), Transform(matrix).to_hexad())
            )
        else:
            val = False
        return val

    def __matmul__(self, matrix):
        # type: (MatrixLike) -> Transform
        """Combine this transform's internal matrix with the given matrix"""
        # Conform the input to a known quantity (and convert if needed)
        other = Transform(matrix)
        # Return a transformation as the combined result
        return Transform(
            (
                self.a * other.a + self.c * other.b,
                self.b * other.a + self.d * other.b,
                self.a * other.c + self.c * other.d,
                self.b * other.c + self.d * other.d,
                self.a * other.e + self.c * other.f + self.e,
                self.b * other.e + self.d * other.f + self.f,
            )
        )

    def __imatmul__(self, matrix):
        # type: (MatrixLike) -> Transform
        """In place multiplication of transform matrices"""
        self.matrix = (self @ matrix).matrix
        if self.callback is not None:
            self.callback(self)
        return self

    def __neg__(self):
        # type: () -> Transform
        """Returns an inverted transformation"""
        det = (self.a * self.d) - (self.c * self.b)
        # invert the rotation/scaling part
        new_a = self.d / det
        new_d = self.a / det
        new_c = -self.c / det
        new_b = -self.b / det
        # invert the translational part
        new_e = -(new_a * self.e + new_c * self.f)
        new_f = -(new_b * self.e + new_d * self.f)
        return Transform((new_a, new_b, new_c, new_d, new_e, new_f))

    def apply_to_point(self, point):
        # type: (VectorLike) -> Vector2d
        """Transform a tuple (X, Y)"""
        if isinstance(point, str):
            raise ValueError(f"Will not transform string '{point}'")
        point = Vector2d(point)
        return Vector2d(
            self.a * point.x + self.c * point.y + self.e,
            self.b * point.x + self.d * point.y + self.f,
        )

    def _is_URT(self, exactly=False):
        # type: (bool) -> bool
        """
        Checks that transformation can be decomposed into product of
        Uniform scale (U), Rotation around origin (R) and translation (T)

        :return: decomposition as U*R*T is possible
        """
        tol = self.absolute_tolerance if not exactly else 0.0
        return (fabs(self.a - self.d) <= tol) and (fabs(self.b + self.c) <= tol)

    def interpolate(self, other, fraction):
        # type: (Transform, float) -> Transform
        """Interpolate with another Transform.

        .. versionadded:: 1.1
        """
        from .tween import TransformInterpolator

        return TransformInterpolator(self, other).interpolate(fraction)


class BoundingInterval:  # pylint: disable=too-few-public-methods
    """A pair of numbers that represent the minimum and maximum values."""

    @overload
    def __init__(self, other=None):
        # type: (Optional[BoundingInterval]) -> None
        pass

    @overload
    def __init__(self, pair):
        # type: (Tuple[float, float]) -> None
        pass

    @overload
    def __init__(self, value):
        # type: (float) -> None
        pass

    @overload
    def __init__(self, x, y):
        # type: (float, float) -> None
        pass

    def __init__(self, x=None, y=None):
        self.x: Union[int, float, Decimal]
        self.y: Union[int, float, Decimal]
        self.minimum: float
        self.maximum: float
        if y is not None:
            if isinstance(x, (int, float, Decimal)) and isinstance(
                y, (int, float, Decimal)
            ):
                self.minimum = x
                self.maximum = y
            else:
                raise ValueError(
                    f"Not a number for scaling: {str((x, y))} "
                    f"({type(x).__name__},{type(y).__name__})"
                )

        else:
            value = x
            if value is None:
                # identity for addition, zero for intersection
                self.minimum, self.maximum = float("+inf"), float("-inf")
            elif isinstance(value, BoundingInterval):
                self.minimum = value.minimum
                self.maximum = value.maximum
            elif isinstance(value, (tuple, list)) and len(value) == 2:
                self.minimum, self.maximum = min(value), max(value)
            elif isinstance(value, (int, float, Decimal)):
                self.minimum = self.maximum = value
            else:
                raise ValueError(
                    f"Not a number for scaling: {str(value)} ({type(value).__name__})"
                )

    def __bool__(self):
        # type: () -> bool
        return isfinite(self.minimum) and isfinite(self.maximum)

    __nonzero__ = __bool__

    def __neg__(self):
        # type: () -> BoundingInterval
        return BoundingInterval((-1 * self.maximum, -1 * self.minimum))

    def __add__(self, other):
        # type: (BoundingInterval) -> BoundingInterval
        """Calculate the bounding interval that covers both given bounding intervals"""
        new = BoundingInterval(self)
        if other is not None:
            new += other
        return new

    def __iadd__(self, other):
        # type: (BoundingInterval) -> BoundingInterval
        other = BoundingInterval(other)
        self.minimum = min((self.minimum, other.minimum))
        self.maximum = max((self.maximum, other.maximum))
        return self

    def __radd__(self, other):
        # type: (BoundingInterval) -> BoundingInterval
        if other is None:
            return BoundingInterval(self)
        return self + other

    def __and__(self, other: BoundingInterval) -> BoundingInterval:
        """Calculate the bounding interval where both given bounding intervals
        overlap"""
        new = BoundingInterval(self)
        if other is not None:
            new &= other
        return new

    def __iand__(self, other):
        # type: (BoundingInterval) -> BoundingInterval
        other = BoundingInterval(other)
        self.minimum = max((self.minimum, other.minimum))
        self.maximum = min((self.maximum, other.maximum))
        if self.minimum > self.maximum:
            self.minimum, self.maximum = float("+inf"), float("-inf")
        return self

    def __rand__(self, other):
        # type: (BoundingInterval) -> BoundingInterval
        if other is None:
            return BoundingInterval(self)
        return self & other

    def __mul__(self, other: float) -> BoundingInterval:
        new = BoundingInterval(self)
        if other is not None:
            new *= other
        return new

    def __imul__(self, other: float) -> BoundingInterval:
        self.minimum *= other
        self.maximum *= other
        return self

    def __iter__(self) -> Generator[float, None, None]:
        yield self.minimum
        yield self.maximum

    def __eq__(self, other) -> bool:
        return tuple(self) == tuple(BoundingInterval(other))

    def __contains__(self, value: float) -> bool:
        return self.minimum <= value <= self.maximum

    def __repr__(self) -> str:
        return f"BoundingInterval({self.minimum}, {self.maximum})"

    @property
    def center(self):
        # type: () -> float
        """Pick the middle of the line"""
        return self.minimum + ((self.maximum - self.minimum) / 2)

    @property
    def size(self):
        # type: () -> float
        """Return the size difference minimum and maximum"""
        return self.maximum - self.minimum


class BoundingBox:  # pylint: disable=too-few-public-methods
    """
    Some functions to compute a rough bbox of a given list of objects.

    BoundingBox(other)
    BoundingBox(x, y)
    BoundingBox((x1, x2), (y1, y2))
    """

    width = property(lambda self: self.x.size)
    height = property(lambda self: self.y.size)
    top = property(lambda self: self.y.minimum)
    left = property(lambda self: self.x.minimum)
    bottom = property(lambda self: self.y.maximum)
    right = property(lambda self: self.x.maximum)
    center_x = property(lambda self: self.x.center)
    center_y = property(lambda self: self.y.center)

    @overload
    def __init__(self, other=None):
        # type: (Optional[BoundingBox]) -> None
        pass

    @overload
    def __init__(self, x, y):
        # type: (BoundingIntervalArgs, BoundingIntervalArgs) -> None
        pass

    def __init__(self, x=None, y=None):
        if y is None:
            if x is None:
                # identity for addition, zero for intersection
                pass
            elif isinstance(x, BoundingBox):
                x, y = x.x, x.y
            else:
                raise ValueError(
                    f"Not a number for scaling: {str(x)} ({type(x).__name__})"
                )
        self.x = BoundingInterval(x)
        self.y = BoundingInterval(y)

    @staticmethod
    def new_xywh(x: float, y: float, width: float, height: float) -> BoundingBox:
        """Create a bounding box using x, y, width and height

        .. versionadded:: 1.2"""
        return BoundingBox((x, x + width), (y, y + height))

    def __bool__(self):
        # type: () -> bool
        return bool(self.x) and bool(self.y)

    __nonzero__ = __bool__

    def __neg__(self):
        # type: () -> BoundingBox
        return BoundingBox(-self.x, -self.y)

    def __add__(self, other):
        # type: (Optional[BoundingBox]) -> BoundingBox
        """Calculate the bounding box that covers both given bounding boxes"""
        new = BoundingBox(self)
        new += BoundingBox(other)
        return new

    def __iadd__(self, other):
        # type: (Optional[BoundingBox]) -> BoundingBox
        other = BoundingBox(other)
        self.x += other.x
        self.y += other.y
        return self

    def __radd__(self, other):
        # type: (Optional[BoundingBox]) -> BoundingBox
        return self + other

    def __and__(self, other):
        # type: (Optional[BoundingBox]) -> BoundingBox
        """Calculate the bounding box where both given bounding boxes overlap"""
        new = BoundingBox(self)
        new &= BoundingBox(other)
        return new

    def __iand__(self, other: Optional[BoundingBox]) -> BoundingBox:
        other = BoundingBox(other)
        self.x = self.x & other.x
        self.y = self.y & other.y
        if not self.x or not self.y:
            self.x, self.y = BoundingInterval(), BoundingInterval()
        return self

    def __rand__(self, other):
        # type: (Optional[BoundingBox]) -> BoundingBox
        return self & other

    def __mul__(self, factor):
        # type: (float) -> BoundingBox
        new = BoundingBox(self)
        new *= factor
        return new

    def __imul__(self, factor):
        # type: (float) -> BoundingBox
        self.x *= factor
        self.y *= factor
        return self

    def __eq__(self, other):
        # type (object) -> bool
        if isinstance(other, BoundingBox):
            return tuple(self) == tuple(other)
        return False

    def __iter__(self) -> Generator[BoundingBox, None, None]:
        yield self.x
        yield self.y

    @property
    def area(self):
        """Return area of the bounding box

        .. versionadded:: 1.2"""
        return self.width * self.height

    @property
    def minimum(self):
        # type: () -> Vector2d
        """Return the minimum x,y coords"""
        return Vector2d(self.x.minimum, self.y.minimum)

    @property
    def maximum(self):
        # type: () -> Vector2d
        """Return the maximum x,y coords"""
        return Vector2d(self.x.maximum, self.y.maximum)

    def __repr__(self):
        # type: () -> str
        return f"BoundingBox({tuple(self.x)},{tuple(self.y)})"

    @property
    def center(self):
        # type: () -> Vector2d
        """Returns the middle of the bounding box"""
        return Vector2d(self.x.center, self.y.center)

    @property
    def size(self):
        """Returns a vector containing width and height of the bounding box

        .. versionadded:: 1.2"""
        return Vector2d(self.x.size, self.y.size)

    def get_anchor(self, xanchor, yanchor, direction=0, selbox=None):
        # type: (str, str, Union[int, str], Optional[BoundingBox]) -> float
        """Calls get_distance with the given anchor options"""
        return self.anchor_distance(
            getattr(self, XAN[xanchor]),
            getattr(self, YAN[yanchor]),
            direction=direction,
            selbox=selbox,
        )

    @staticmethod
    def anchor_distance(
        x: float,
        y: float,
        direction: Union[int, str] = 0,
        selbox: Optional[BoundingBox] = None,
    ) -> float:
        """Using the x,y returns a single sortable value based on direction and angle

        Args:
            x (float): input x coordinate
            y (float): input y coordinate
            direction (Union[int, str], optional): int/float (custom angle),
                tb/bt (top/bottom), lr/rl (left/right), ri/ro (radial). Defaults to 0.
            selbox (Optional[BoundingBox], optional): The bounding box of the whole
                selection for radial anchors. Defaults to None.

        Raises:
            ValueError: if radial distance is requested without the optional selbox
                parameter.

        Returns:
            float: the anchor distance with respect to the direction.
        """

        rot = 0.0
        if isinstance(direction, (int, float)):  # Angle
            if direction not in CUSTOM_DIRECTION:
                return hypot(x, y) * (cos(radians(-direction) - atan2(y, x)))
            direction = CUSTOM_DIRECTION[direction]

        if direction in ("ro", "ri"):
            if selbox is None:
                raise ValueError(
                    "Radial distance not available without selection bounding box"
                )
            rot = hypot(selbox.x.center - x, selbox.y.center - y)

        return [y, -y, x, -x, rot, -rot][DIRECTION.index(direction)]

    def resize(self, delta_x: float, delta_y: float = None) -> BoundingBox:
        """Enlarges / shrinks a bounding box by a constant value. If only delta_x
        is given, each side is moved by the same amount; if delta_y is given,
        different deltas are applied to horizontal and vertical intervals.

        .. versionadded:: 1.2"""
        delta_y = delta_y or delta_x
        return BoundingBox(
            (self.x.minimum - delta_x, self.x.maximum + delta_x),
            (self.y.minimum - delta_y, self.y.maximum + delta_y),
        )


class DirectedLineSegment:
    """
    A directed line segment

    DirectedLineSegment(((x0, y0), (x1, y1)))
    """

    start = Vector2d()  # start point of segment
    end = Vector2d()  # end point of segment

    x0 = property(lambda self: self.start.x)  # pylint: disable=invalid-name
    y0 = property(lambda self: self.start.y)  # pylint: disable=invalid-name
    x1 = property(lambda self: self.end.x)
    y1 = property(lambda self: self.end.y)
    dx = property(lambda self: self.vector.x)  # pylint: disable=invalid-name
    dy = property(lambda self: self.vector.y)  # pylint: disable=invalid-name

    @overload
    def __init__(self):
        # type: () -> None
        pass

    @overload
    def __init__(self, other):
        # type: (DirectedLineSegment) -> None
        pass

    @overload
    def __init__(self, start, end):
        # type: (VectorLike, VectorLike) -> None
        pass

    def __init__(self, *args):
        if not args:  # overload 0
            start, end = Vector2d(), Vector2d()
        elif len(args) == 1:  # overload 1
            (other,) = args
            start, end = other.start, other.end
        elif len(args) == 2:  # overload 2
            start, end = args
        else:
            raise ValueError(f"DirectedLineSegment() can't be constructed from {args}")

        self.start = Vector2d(start)
        self.end = Vector2d(end)

    def __eq__(self, other):
        # type: (object) -> bool
        if isinstance(other, (tuple, DirectedLineSegment)):
            return tuple(self) == tuple(other)
        return False

    def __iter__(self):
        # type: () -> Generator[DirectedLineSegment, None, None]
        yield self.x0
        yield self.x1
        yield self.y0
        yield self.y1

    @property
    def vector(self):
        # type: () -> Vector2d
        """The vector of the directed line segment.

        The vector of the directed line segment represents the length
        and direction of segment, but not the starting point.

        .. versionadded:: 1.1
        """
        return self.end - self.start

    @property
    def length(self):
        # type: () -> float
        """Get the length of the line segment"""
        return self.vector.length

    @property
    def angle(self):
        # type: () -> float
        """Get the angle of the line created by this segment"""
        return atan2(self.dy, self.dx)

    def distance_to_point(self, x, y):
        # type: (float, float) -> Union[DirectedLineSegment, Optional[float]]
        """Get the distance to the given point (x, y)"""
        segment2 = DirectedLineSegment(self.start, (x, y))
        dot2 = segment2.dot(self)
        if dot2 <= 0:
            return DirectedLineSegment((x, y), self.start).length
        if self.dot(self) <= dot2:
            return DirectedLineSegment((x, y), self.end).length
        return self.perp_distance(x, y)

    def perp_distance(self, x, y):
        # type: (float, float) -> Optional[float]
        """Perpendicular distance to the given point"""
        if self.length == 0:
            return None
        return fabs((self.dx * (self.y0 - y)) - ((self.x0 - x) * self.dy)) / self.length

    def dot(self, other):
        # type: (DirectedLineSegment) -> float
        """Get the dot product with the segment with another"""
        return self.vector.dot(other.vector)

    def point_at_ratio(self, ratio):
        # type: (float) -> Tuple[float, float]
        """Get the point at the given ratio along the line"""
        return self.x0 + ratio * self.dx, self.y0 + ratio * self.dy

    def point_at_length(self, length):
        # type: (float) -> Tuple[float, float]
        """Get the point as the length along the line"""
        return self.point_at_ratio(length / self.length)

    def parallel(self, x, y):
        # type: (float, float) -> DirectedLineSegment
        """Create parallel Segment"""
        return DirectedLineSegment((x + self.dx, y + self.dy), (x, y))

    def intersect(self, other):
        # type: (DirectedLineSegment) -> Optional[Vector2d]
        """Get the intersection between two segments"""
        other = DirectedLineSegment(other)
        denom = self.vector.cross(other.vector)
        num = other.vector.cross(self.start - other.start)

        if denom != 0:
            return Vector2d(self.point_at_ratio(num / denom))
        return None

    def __repr__(self):
        # type: () -> str
        return f"DirectedLineSegment(({self.start}), ({self.end}))"


def cubic_extrema(py0, py1, py2, py3):
    # type: (float, float, float, float) -> Tuple[float, float]
    """Returns the extreme value, given a set of bezier coordinates"""

    atol = 1e-9
    cmin, cmax = min(py0, py3), max(py0, py3)
    pd1 = py1 - py0
    pd2 = py2 - py1
    pd3 = py3 - py2

    def _is_bigger(point):
        if 0 < point < 1:
            pyx = (
                py0 * (1 - point) * (1 - point) * (1 - point)
                + 3 * py1 * point * (1 - point) * (1 - point)
                + 3 * py2 * point * point * (1 - point)
                + py3 * point * point * point
            )
            return min(cmin, pyx), max(cmax, pyx)
        return cmin, cmax

    if fabs(pd1 - 2 * pd2 + pd3) > atol:
        if pd2 * pd2 > pd1 * pd3:
            pds = sqrt(pd2 * pd2 - pd1 * pd3)
            cmin, cmax = _is_bigger((pd1 - pd2 + pds) / (pd1 - 2 * pd2 + pd3))
            cmin, cmax = _is_bigger((pd1 - pd2 - pds) / (pd1 - 2 * pd2 + pd3))

    elif fabs(pd2 - pd1) > atol:
        cmin, cmax = _is_bigger(-pd1 / (2 * (pd2 - pd1)))

    return cmin, cmax


def quadratic_extrema(py0, py1, py2):
    # type: (float, float, float) -> Tuple[float, float]
    """Returns the extreme value, given a set of quadratic bezier coordinates"""
    atol = 1e-9
    cmin, cmax = min(py0, py2), max(py0, py2)

    def _is_bigger(point):
        if 0 < point < 1:
            pyx = (
                py0 * (1 - point) * (1 - point)
                + 2 * py1 * point * (1 - point)
                + py2 * point * point
            )
            return min(cmin, pyx), max(cmax, pyx)
        return cmin, cmax

    if fabs(py0 + py2 - 2 * py1) > atol:
        cmin, cmax = _is_bigger((py0 - py1) / (py0 + py2 - 2 * py1))

    return cmin, cmax



# coding=utf-8
#
# Copyright (C) 2011 Karlisson Bezerra <contact@hacktoon.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Element parsing and context for ink2canvas extensions
"""

from __future__ import unicode_literals
import math

import inkex
from inkex.localization import inkex_gettext as _


# pylint: disable=missing-function-docstring, missing-class-docstring
# pylint: disable=too-few-public-methods


class Element:
    """Base Element"""

    def __init__(self, node):
        self.node = node

    def attr(self, val):
        """Get attribute"""
        try:
            attr = float(self.node.get(val))
        except (ValueError, TypeError, AttributeError):
            attr = self.node.get(val)
        return attr


class GradientDef(Element):
    def __init__(self, node, stops):
        super().__init__(node)
        self.stops = stops


class LinearGradientDef(GradientDef):
    def get_data(self):
        # pylint: disable=unused-variable
        x1 = self.attr("x1")
        y1 = self.attr("y1")
        x2 = self.attr("x2")
        y2 = self.attr("y2")
        # self.create_linear_gradient(href, x1, y1, x2, y2)

    def draw(self):
        pass


class RadialGradientDef(GradientDef):
    def get_data(self):
        # pylint: disable=unused-variable
        cx = self.attr("cx")
        cy = self.attr("cy")
        r = self.attr("r")
        # self.create_radial_gradient(href, cx, cy, r, cx, cy, r)

    def draw(self):
        pass


class AbstractShape(Element):
    def __init__(self, command, node, ctx):
        super().__init__(node)
        self.command = command
        self.ctx = ctx
        self.gradient = None

    def get_data(self):  # pylint: disable=no-self-use
        return None

    def get_style(self):
        return self.node.specified_style()

    def set_style(self, style):
        """Translates style properties names into method calls"""
        self.ctx.style = style
        for key in style:
            method = "set_" + "_".join(key.split("-"))
            if hasattr(self.ctx, method) and style[key] != "none":
                getattr(self.ctx, method)(style[key])
        # saves style to compare in next iteration
        if (
            hasattr(self.ctx, "style_cache")
            and "opacity" not in style
            and self.ctx.style_cache("opacity") != style("opacity")
        ):
            self.ctx.set_opacity(
                style("opacity")
            )  # opacity is kept in memory, need to reset
        self.ctx.style_cache = style

    def has_transform(self):
        return bool(self.attr("transform"))

    def get_transform(self):
        return self.node.transform.to_hexad()

    def has_gradient(self):
        style = self.get_style()
        fill = style("fill")
        return fill is not None and isinstance(fill, inkex.Gradient)

    def get_gradient_href(self):
        style = self.get_style()
        return style("fill").get_id()

    def has_clip(self):
        return bool(self.attr("clip-path"))

    def start(self, gradient):
        self.gradient = gradient
        self.ctx.write("\n// #%s" % self.attr("id"))
        if self.has_transform() or self.has_clip():
            self.ctx.save()

    def draw(self):
        data = self.get_data()  # pylint: disable=assignment-from-none
        style = self.get_style()
        self.ctx.begin_path()
        if self.has_transform():
            trans_matrix = self.get_transform()
            self.ctx.transform(*trans_matrix)  # unpacks argument list
        if self.has_gradient():
            self.gradient.draw()
        self.set_style(style)
        # unpacks "data" in parameters to given method
        getattr(self.ctx, self.command)(*data)
        self.ctx.finish_path()

    def end(self):
        if self.has_transform() or self.has_clip():
            self.ctx.restore()


class G(AbstractShape):  # pylint: disable=invalid-name
    def draw(self):
        # get layer label, if exists
        if self.has_transform():
            trans_matrix = self.get_transform()
            self.ctx.transform(*trans_matrix)


class Rect(AbstractShape):
    def get_data(self):
        x = self.attr("x")
        y = self.attr("y")
        width = self.attr("width")
        height = self.attr("height")
        rx = self.attr("rx") or 0
        ry = self.attr("ry") or 0
        return x, y, width, height, rx, ry


class Circle(AbstractShape):
    def __init__(self, command, node, ctx):
        AbstractShape.__init__(self, command, node, ctx)
        self.command = "arc"

    def get_data(self):

        cx = self.attr("cx")
        cy = self.attr("cy")
        r = self.attr("r")
        return cx, cy, r, 0, math.pi * 2, True


class Ellipse(AbstractShape):
    def get_data(self):
        cx = self.attr("cx")
        cy = self.attr("cy")
        rx = self.attr("rx")
        ry = self.attr("ry")
        return cx, cy, rx, ry

    def draw(self):
        cx, cy, rx, ry = self.get_data()
        style = self.get_style()
        self.ctx.begin_path()
        if self.has_transform():
            trans_matrix = self.get_transform()
            self.ctx.transform(*trans_matrix)  # unpacks argument list
        self.set_style(style)

        kappa = 4 * ((math.sqrt(2) - 1) / 3)
        self.ctx.move_to(cx, cy - ry)
        self.ctx.bezier_curve_to(
            cx + (kappa * rx), cy - ry, cx + rx, cy - (kappa * ry), cx + rx, cy
        )
        self.ctx.bezier_curve_to(
            cx + rx, cy + (kappa * ry), cx + (kappa * rx), cy + ry, cx, cy + ry
        )
        self.ctx.bezier_curve_to(
            cx - (kappa * rx), cy + ry, cx - rx, cy + (kappa * ry), cx - rx, cy
        )
        self.ctx.bezier_curve_to(
            cx - rx, cy - (kappa * ry), cx - (kappa * rx), cy - ry, cx, cy - ry
        )
        self.ctx.finish_path()


class Path(AbstractShape):
    def __init__(self, command, node, ctx):
        AbstractShape.__init__(self, command, node, ctx)
        self.current_position = 0, 0

    def path_move_to(self, data):
        self.ctx.move_to(data[0], data[1])
        self.current_position = data[0], data[1]

    def path_line_to(self, data):
        self.ctx.line_to(data[0], data[1])
        self.current_position = data[0], data[1]

    def path_curve_to(self, data):
        x1, y1, x2, y2 = data[0], data[1], data[2], data[3]
        x, y = data[4], data[5]
        self.ctx.bezier_curve_to(x1, y1, x2, y2, x, y)
        self.current_position = x, y

    def path_close(self, data):  # pylint: disable=unused-argument
        self.ctx.close_path()

    def draw(self):
        """Gets the node type and calls the given method"""
        style = self.get_style()
        self.ctx.begin_path()
        if self.has_transform():
            trans_matrix = self.get_transform()
            self.ctx.transform(*trans_matrix)  # unpacks argument list
        self.set_style(style)

        # Draws path commands
        path_command = {
            "M": self.path_move_to,
            "L": self.path_line_to,
            "C": self.path_curve_to,
            "Z": self.path_close,
        }
        # Make sure we only have Lines and curves (no arcs etc)
        for comm, data in self.node.path.to_superpath().to_path().to_arrays():
            if comm in path_command:
                path_command[comm](data)

        self.ctx.finish_path()


class Line(Path):
    def get_data(self):
        x1 = self.attr("x1")
        y1 = self.attr("y1")
        x2 = self.attr("x2")
        y2 = self.attr("y2")
        return ("M", (x1, y1)), ("L", (x2, y2))


class Polygon(Path):
    def get_data(self):
        points = self.attr("points").strip().split(" ")
        points = map(lambda x: x.split(","), points)
        comm = []
        for point in points:  # creating path command similar
            point = list(map(float, point))
            comm.append(["L", point])
        comm[0][0] = "M"  # first command must be a 'M' => moveTo
        return comm


class Polyline(Polygon):
    pass


class Text(AbstractShape):
    def text_helper(self, tspan):
        if tspan is not None:
            return tspan.text
        for ts_cur in tspan:
            return ts_cur.text + self.text_helper(ts_cur) + ts_cur.tail

    def set_text_style(self, style):
        keys = ("font-style", "font-weight", "font-size", "font-family")
        text = []
        for key in keys:
            if key in style:
                text.append(style[key])
        self.ctx.set_font(" ".join(text))

    def get_data(self):
        x = self.attr("x")
        y = self.attr("y")
        return x, y

    def draw(self):
        for tspan in self.node:
            if isinstance(tspan, inkex.TextPath):
                raise ValueError(_("TextPath elements are not supported"))
        style = self.get_style()
        if self.has_transform():
            trans_matrix = self.get_transform()
            self.ctx.transform(*trans_matrix)  # unpacks argument list
        self.set_style(style)
        self.set_text_style(style)

        for tspan in self.node:
            text = self.text_helper(tspan)
            cur_x = float(tspan.get("x").split()[0])
            cur_y = float(tspan.get("y").split()[0])
            self.ctx.fill_text(text, cur_x, cur_y)



#!/usr/bin/env python
#
# Copyright 2008, 2009 Hannes Hochreiner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.
#
"""Add a video to the slideshow"""

import re
from copy import deepcopy

import inkex
from jessyink_install import JessyInkMixin, _


class Video(JessyInkMixin, inkex.EffectExtension):
    """Add jessyink video"""

    def add_arguments(self, pars):
        self.arg_parser.add_argument("--tab", dest="what")

    def effect(self):
        self.is_installed()
        # Check version.
        base_view = self.svg.xpath("//sodipodi:namedview[@id='base']")
        if base_view is None:
            raise inkex.AbortExtension(
                _(
                    "Could not obtain the selected layer for inclusion of the video element."
                )
            )

        layer = self.svg.get_current_layer()
        if layer is None:
            raise inkex.AbortExtension(
                _(
                    "Could not obtain the selected layer for inclusion of the video element.\n\n"
                )
            )

        template = inkex.load_svg(self.get_resource("jessyink_video.svg"))
        root = template.getroot()

        elem = layer.add(
            root.getElement("//svg:g[@jessyink:element='core.video']").copy()
        )
        node_dict = find_internal_links(elem, root, {})
        delete_ids(elem)

        for node in node_dict.values():
            elem.insert(0, node)

        for node in node_dict.values():
            node.set_id(self.svg.get_unique_id("jessyink.core.video"), backlinks=True)


def find_internal_links(node, svg, node_dict):
    """Get all clone links and css links"""
    for entry in re.findall(rb"url\(#.*\)", node.tostring()):
        entry = entry.decode()
        link_id = entry[5 : len(entry) - 1]

        if link_id not in node_dict:
            node_dict[link_id] = deepcopy(svg.getElementById(link_id))
            find_internal_links(node_dict[link_id], svg, node_dict)

    for entry in node.iter():
        if entry.get("xlink:href"):
            link_id = entry.get("xlink:href")
            if link_id not in node_dict:
                node_dict[link_id] = deepcopy(svg.getElementById(link_id))
                find_internal_links(node_dict[link_id], svg, node_dict)

    return node_dict


def delete_ids(node):
    """Delete ids in the given node's children"""
    for entry in node.iter():
        if "id" in entry.attrib:
            del entry.attrib["id"]


if __name__ == "__main__":
    Video().run()



#!/usr/bin/env python
"""Flip upper to lower and lower to upper cases"""

import inkex


class FlipCase(inkex.TextExtension):
    """Change the case, cHANGE THE CASE"""

    @staticmethod
    def map_char(char):
        return char.upper() if char.islower() else char.lower()


if __name__ == "__main__":
    FlipCase().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2012 Jabiertxo Arraiza, jabier.arraiza@marker.es
# Copyright (C) 2016 su_v, <suv-sf@users.sf.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""Document information"""

import inkex

from inkex.localization import (
    inkex_gettext as _,
    inkex_fgettext as _f,
    inkex_pgettext as pgettext,
)
import inkex.localization


class DocInfo(inkex.EffectExtension):
    """Show document information"""

    def effect(self):
        namedview = self.svg.namedview
        self.msg(pgettext("Docinfo Extension", ":::SVG document related info:::"))
        self.msg(
            _f(
                "version: {}",
                self.svg.get("inkscape:version", _("New Document (unsaved)")),
            )
        )
        self.msg(
            pgettext("Docinfo Extension", "width: {}").format(self.svg.viewport_width)
        )
        self.msg(
            pgettext("Docinfo Extension", "height: {}").format(self.svg.viewport_height)
        )
        self.msg(
            pgettext("Docinfo Extension", "viewbox: {}").format(
                str(self.svg.get_viewbox())
            )
        )
        self.msg(
            pgettext("Docinfo Extension", "document-units: {}").format(
                namedview.get("inkscape:document-units", "None")
            )
        )
        self.msg(
            pgettext("Docinfo Extension", "units: ") + namedview.get("units", "None")
        )
        self.msg(_f("Document has {} guides", len(namedview.get_guides())))
        for i, grid in enumerate(namedview.findall("inkscape:grid")):
            self.msg(_f("Grid number {}: Units: {}", i + 1, grid.get("units", "None")))
        if len(namedview.get_pages()) > 1:
            self.msg(_f("Document has {} pages", len(namedview.get_pages())))
            for i, page in enumerate(namedview.get_pages()):
                self.msg(
                    _f(
                        "Page number {}: x: {} y: {} width: {} height: {}",
                        i + 1,
                        page.get("x"),
                        page.get("y"),
                        page.get("width"),
                        page.get("height"),
                    )
                )
        else:
            self.msg(_("This is a single page document."))


if __name__ == "__main__":
    DocInfo().run()



"""
Useful decorators for tests.
"""
import pytest
from inkex.command import is_inkscape_available

requires_inkscape = pytest.mark.skipif(  # pylint: disable=invalid-name
    not is_inkscape_available(), reason="Test requires inkscape, but it's not available"
)



# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#               2020 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Module for interpolating attributes and styles

.. versionchanged:: 1.2
    Rewritten in inkex 1.2 in an object-oriented structure to support more attributes.
"""
from bisect import bisect_left
import abc
import copy

from .styles import Style
from .elements._filters import LinearGradient, RadialGradient, Stop
from .transforms import Transform
from .colors import Color
from .units import convert_unit, parse_unit, render_unit
from .bezier import bezlenapprx, cspbezsplit, cspbezsplitatlength, csplength
from .paths import Path, CubicSuperPath
from .elements import SvgDocumentElement
from .utils import FragmentError


try:
    from typing import Tuple, TypeVar

    Value = TypeVar("Value")
    Number = TypeVar("Number", int, float)
except ImportError:
    pass


def interpcoord(coord_a: Number, coord_b: Number, time: float):
    """Interpolate single coordinate by the amount of time"""
    return ValueInterpolator(coord_a, coord_b).interpolate(time)


def interppoints(point1, point2, time):
    # type: (Tuple[float, float], Tuple[float, float], float) -> Tuple[float, float]
    """Interpolate coordinate points by amount of time"""
    return ArrayInterpolator(point1, point2).interpolate(time)


class AttributeInterpolator(abc.ABC):
    """Interpolate between attributes"""

    def __init__(self, start_value, end_value):
        self.start_value = start_value
        self.end_value = end_value

    @staticmethod
    def best_style(node):
        """Gets the best possible approximation to a node's style. For nodes inside the
        element tree of an SVG file, stylesheets defined in the defs of that file can be
        taken into account. This should be the case for input elements, but is not
        required - in that case, only the local inline style is used.

        During the interpolation process, some nodes are created temporarily, such as
        plain gradients of a single color to allow solid<->gradient interpolation. These
        are not attached to the document tree and therefore have no root. Since the only
        style relevant for them is the inline style, it is acceptable to fallback to it.

        Args:
            node (BaseElement): The node to get the best approximated style of

        Returns:
            Style: If the node is rooted, the CSS specified style. Else, the inline
            style."""
        try:
            return node.specified_style()
        except FragmentError:
            return node.style

    @staticmethod
    def create_from_attribute(snode, enode, attribute, method=None):
        """Creates an interpolator for an attribute. Currently, only path, transform and
        style attributes are supported

        Args:
            snode (BaseElement): start element
            enode (BaseElement): end element
            attribute (str): attribute name (for styles, starting with "style/")
            method (AttributeInterpolator, optional): (currently only used for paths).
                Specifies a method used to interpolate the attribute. Defaults to None.

        Raises:
            ValueError: if an attribute is passed that is not a style, path or transform
                attribute

        Returns:
            AttributeInterpolator: an interpolator whose type depends on attribute.
        """
        if attribute in Style.color_props:
            return StyleInterpolator.create_from_fill_stroke(snode, enode, attribute)
        if attribute == "d":
            if method is None:
                method = FirstNodesInterpolator
            return method(snode.path, enode.path)
        if attribute == "style":
            return StyleInterpolator(snode, enode)
        if attribute.startswith("style/"):
            return StyleInterpolator.create(snode, enode, attribute[6:])
        if attribute == "transform":
            return TransformInterpolator(snode.transform, enode.transform)
        if method is not None:
            return method(snode.get(attribute), enode.get(attribute))
        raise ValueError("only path and style attributes are supported")

    @abc.abstractmethod
    def interpolate(self, time=0):
        """Interpolation method, needs to be implemented by subclasses"""
        return


class StyleInterpolator(AttributeInterpolator):
    """Class to interpolate styles"""

    def __init__(self, start_value, end_value):
        super().__init__(start_value, end_value)
        self.interpolators = {}
        # some keys are always processed in a certain order, these provide alternative
        # interpolation routes if e.g. Color<->none is interpolated
        all_keys = list(
            dict.fromkeys(
                ["fill", "stroke", "fill-opacity", "stroke-opacity", "stroke-width"]
                + list(self.best_style(start_value).keys())
                + list(self.best_style(end_value).keys())
            )
        )
        for attr in all_keys:
            sstyle = self.best_style(start_value)
            estyle = self.best_style(end_value)
            if attr not in sstyle and attr not in estyle:
                continue
            try:
                interp = StyleInterpolator.create(
                    self.start_value, self.end_value, attr
                )
                self.interpolators[attr] = interp
            except ValueError:
                # no interpolation method known for this attribute
                pass

    @staticmethod
    def create(snode, enode, attribute):
        """Creates an Interpolator for a given style attribute, depending on its type:

            - Color properties (such as fill, stroke) -> :class:`ColorInterpolator`,
              :class:`GradientInterpolator` ect.
            - Unit properties -> :class:`UnitValueInterpolator`
            - other properties -> :class:`ValueInterpolator`

        Args:
            snode (BaseElement): start element
            enode (BaseElement): end element
            attribute (str): attribute to interpolate

        Raises:
            ValueError: if the attribute is not in any of the lists

        Returns:
            AttributeInterpolator: an interpolator object whose type depends on the
            attribute.
        """
        if attribute in Style.color_props:
            return StyleInterpolator.create_from_fill_stroke(snode, enode, attribute)

        if attribute in Style.unit_props:
            return UnitValueInterpolator(
                AttributeInterpolator.best_style(snode)(attribute),
                AttributeInterpolator.best_style(enode)(attribute),
            )

        if attribute in Style.opacity_props:
            return ValueInterpolator(
                AttributeInterpolator.best_style(snode)(attribute),
                AttributeInterpolator.best_style(enode)(attribute),
            )

        raise ValueError("Unknown attribute")

    @staticmethod
    def create_from_fill_stroke(snode, enode, attribute):
        """Creates an Interpolator for a given color-like attribute

        Args:
            snode (BaseElement): start element
            enode (BaseElement): end element
            attribute (str): attribute to interpolate

        Raises:
            ValueError: if the attribute is not color-like
            ValueError: if the attribute is unset on both start and end style

        Returns:
            AttributeInterpolator: an interpolator object whose type depends on the
            attribute.
        """
        if attribute not in Style.color_props:
            raise ValueError("attribute must be a color property")

        sstyle = AttributeInterpolator.best_style(snode)
        estyle = AttributeInterpolator.best_style(enode)

        styles = [[snode, sstyle], [enode, estyle]]
        for (cur, curstyle) in styles:
            if curstyle(attribute) is None:
                cur.style[attribute + "-opacity"] = 0.0
                if attribute == "stroke":
                    cur.style["stroke-width"] = 0.0

        # check if style is none, unset or a color
        if isinstance(
            sstyle(attribute), (LinearGradient, RadialGradient)
        ) or isinstance(estyle(attribute), (LinearGradient, RadialGradient)):
            # if one of the two styles is a gradient, use gradient interpolation.
            try:
                return GradientInterpolator.create(snode, enode, attribute)
            except ValueError:
                # different gradient types, just duplicate the first
                return TrivialInterpolator(sstyle(attribute))
        if sstyle(attribute) is None and estyle(attribute) is None:
            return TrivialInterpolator("none")
        return ColorInterpolator.create(sstyle, estyle, attribute)

    def interpolate(self, time=0):
        """Interpolates a style using the interpolators set in self.interpolators

        Args:
            time (int, optional): Interpolation position. If 0, start_value is returned,
                if 1, end_value is returned. Defaults to 0.

        Returns:
            inkex.Style: interpolated style
        """
        style = Style()
        for prop, interp in self.interpolators.items():
            style[prop] = interp.interpolate(time)
        return style


class TrivialInterpolator(AttributeInterpolator):
    """Trivial interpolator, returns value for every time"""

    def __init__(self, value):
        super().__init__(value, value)

    def interpolate(self, time=0):
        return self.start_value


class ValueInterpolator(AttributeInterpolator):
    """Class for interpolation of a single value"""

    def __init__(self, start_value=0, end_value=0):
        super().__init__(float(start_value), float(end_value))

    def interpolate(self, time=0):
        """(Linearly) interpolates a value

        Args:
            time (int, optional): Interpolation position. If 0, start_value is returned,
                if 1, end_value is returned. Defaults to 0.

        Returns:
            int: interpolated value
        """
        return self.start_value + ((self.end_value - self.start_value) * time)


class UnitValueInterpolator(ValueInterpolator):
    """Class for interpolation of a value with unit"""

    def __init__(self, start_value=0, end_value=0):
        start_val, start_unit = parse_unit(start_value)
        end_val = convert_unit(end_value, start_unit)
        super().__init__(start_val, end_val)
        self.unit = start_unit

    def interpolate(self, time=0):
        return render_unit(super().interpolate(time), self.unit)


class ArrayInterpolator(AttributeInterpolator):
    """Interpolates array-like objects element-wise, e.g. color, transform,
    coordinate"""

    def __init__(self, start_value, end_value):
        super().__init__(start_value, end_value)
        self.interpolators = [
            ValueInterpolator(cur, other)
            for (cur, other) in zip(start_value, end_value)
        ]

    def interpolate(self, time=0):
        """Interpolates an array element-wise

        Args:
            time (int, optional): [description]. Defaults to 0.

        Returns:
            List: interpolated array
        """
        return [interp.interpolate(time) for interp in self.interpolators]


class TransformInterpolator(ArrayInterpolator):
    """Class for interpolation of transforms"""

    def __init__(self, start_value=Transform(), end_value=Transform()):
        """Creates a transform interpolator.

        Args:
            start_value (inkex.Transform, optional): start transform. Defaults to
                inkex.Transform().
            end_value (inkex.Transform, optional): end transform. Defaults to
                inkex.Transform().
        """
        super().__init__(start_value.to_hexad(), end_value.to_hexad())

    def interpolate(self, time=0):
        """Interpolates a transform by interpolating each item in the transform hexad
        separately.

        Args:
            time (int, optional): Interpolation position. If 0, start_value is returned,
                if 1, end_value is returned. Defaults to 0.

        Returns:
            Transform: interpolated transform
        """
        return Transform(super().interpolate(time))


class ColorInterpolator(ArrayInterpolator):
    """Class for color interpolation"""

    @staticmethod
    def create(sst, est, attribute):
        """Creates a ColorInterpolator for either Fill or stroke, depending on the
        attribute.

        Args:
            sst (Style): Start style
            est (Style): End style
            attribute (string): either fill or stroke

        Raises:
            ValueError: if none of the start or end style is a color.

        Returns:
            ColorInterpolator: A ColorInterpolator object
        """
        styles = [sst, est]
        for cur, other in zip(styles, reversed(styles)):
            if not isinstance(cur(attribute), Color) or cur(attribute) is None:
                cur[attribute] = other(attribute)
        this = ColorInterpolator(
            Color(styles[0](attribute)), Color(styles[1](attribute))
        )
        if this is None:
            raise ValueError("One of the two attribute needs to be a plain color")
        return this

    def __init__(self, start_value=Color("#000000"), end_value=Color("#000000")):
        super().__init__(start_value, end_value)

    def interpolate(self, time=0):
        """Interpolates a color by interpolating its r, g, b, a channels separately.

        Args:
            time (int, optional): Interpolation position. If 0, start_value is returned,
                if 1, end_value is returned. Defaults to 0.

        Returns:
            Color: interpolated color
        """
        return Color(list(map(int, super().interpolate(time))))


class GradientInterpolator(AttributeInterpolator):
    """Base class for Gradient Interpolation"""

    def __init__(self, start_value, end_value, svg=None):
        super().__init__(start_value, end_value)
        self.svg = svg
        # If one of the styles is empty, set it to the gradient of the other
        if start_value is None:
            self.start_value = end_value
        if end_value is None:
            self.end_value = start_value
        self.transform_interpolator = TransformInterpolator(
            self.start_value.gradientTransform, self.end_value.gradientTransform
        )
        self.orientation_interpolator = {
            attr: UnitValueInterpolator(
                self.start_value.get(attr), self.end_value.get(attr)
            )
            for attr in self.start_value.orientation_attributes
            if self.start_value.get(attr) is not None
            and self.end_value.get(attr) is not None
        }
        if not (
            self.start_value.href is not None
            and self.start_value.href is self.end_value.href
        ):
            # the gradient link to different stops, interpolate between them
            # add both start and end offsets, then take distict
            newoffsets = sorted(
                list(set(self.start_value.stop_offsets + self.end_value.stop_offsets))
            )

            def func(start, end, time):
                return StopInterpolator(start, end).interpolate(time)

            sstops = GradientInterpolator.interpolate_linear_list(
                self.start_value.stop_offsets,
                list(self.start_value.stops),
                newoffsets,
                func,
            )
            ostops = GradientInterpolator.interpolate_linear_list(
                self.end_value.stop_offsets,
                list(self.end_value.stops),
                newoffsets,
                func,
            )
            self.newstop_interpolator = [
                StopInterpolator(s1, s2) for s1, s2 in zip(sstops, ostops)
            ]
        else:
            self.newstop_interpolator = None

    @staticmethod
    def create(snode, enode, attribute):
        """Creates a `GradientInterpolator` for either fill or stroke, depending on
        attribute.

        Cases: (A, B) -> Interpolator

          - Linear Gradient, Linear Gradient -> LinearGradientInterpolator
          - Color or None, Linear Gradient -> LinearGradientInterpolator
          - Radial Gradient, Radial Gradient -> RadialGradientInterpolator
          - Color or None, Radial Gradient -> RadialGradientInterpolator
          - Radial Gradient, Linear Gradient -> ValueError
          - Color or None, Color or None -> ValueError

        Args:
            snode (BaseElement): start element
            enode (BaseElement): end element
            attribute (string): either fill or stroke

        Raises:
            ValueError: if none of the styles are a gradient or if they are gradients
                of different types

        Returns:
            GradientInterpolator: an Interpolator object
        """
        interpolator = None
        gradienttype = None
        # first find out which type of interpolator we need
        sstyle = AttributeInterpolator.best_style(snode)
        estyle = AttributeInterpolator.best_style(enode)
        for cur in [sstyle, estyle]:
            curgrad = None
            if isinstance(cur(attribute), (LinearGradient, RadialGradient)):
                curgrad = cur(attribute)
            for gradtype, interp in [
                [LinearGradient, LinearGradientInterpolator],
                [RadialGradient, RadialGradientInterpolator],
            ]:
                if curgrad is not None and isinstance(curgrad, gradtype):
                    if interpolator is None:
                        interpolator = interp
                        gradienttype = gradtype
                    if not (interp == interpolator):
                        raise ValueError("Gradient types don't match")
        # If one of the styles is empty, set it to the gradient of the other, but with
        # zero opacity (and stroke-width for strokes)
        # If one of the styles is a plain color, replace it by a gradient with a single
        # stop
        iterator = [[snode, gradienttype(), enode], [enode, gradienttype(), snode]]
        for index in [0, 1]:
            curstyle = AttributeInterpolator.best_style(iterator[index][0])
            value = curstyle(attribute)
            if value is None:
                # if the attribute of one of the two ends is unset, set the opacity to
                # zero.
                iterator[index][0].style[attribute + "-opacity"] = 0.0
                if attribute == "stroke":
                    iterator[index][0].style["stroke-width"] = 0.0
            if isinstance(value, Color):
                # if the attribute of one of the two ends is a color, convert it to a
                # one-stop gradient. Type depends on the type of the other gradient.
                interpolator.initialize_position(
                    iterator[index][1], iterator[index][0].bounding_box()
                )
                stop = Stop()
                stop.style = Style()
                stop.style["stop-color"] = value
                stop.offset = 0
                iterator[index][1].add(stop)
                stop = Stop()
                stop.style = Style()
                stop.style["stop-color"] = value
                stop.offset = 1
                iterator[index][1].add(stop)
            else:
                iterator[index][1] = value  # is a gradient
        if interpolator is None:
            raise ValueError("None of the two styles is a gradient")
        if interpolator in [LinearGradientInterpolator, RadialGradientInterpolator]:
            return interpolator(iterator[0][1], iterator[1][1], snode)
        return interpolator(iterator[0][1], iterator[1][1])

    @staticmethod
    def interpolate_linear_list(positions, values, newpositions, func):
        """Interpolates a list of values given at n positions to the best approximation
        at m newpositions.

        >>>
            |
            |         x
            |  x
            _________________
               pq  q  p   q
            (x denotes function values, p: positions, q: newpositions)
            A function may be given to interpolate between given values.

        Args:
            positions (list[number-like]): position of current function values
            values (list[Type]): list of arbitrary type,
                ``len(values) == len(positions)``
            newpositions (list[number-like]): position of interpolated values
            func (Callable[[Type, Type, float], Type]): Function to interpolate between
                values

        Returns:
            list[Type]: interpolated function values at positions
        """
        newvalues = []
        positions = list(map(float, positions))
        newpositions = list(map(float, newpositions))
        for pos in newpositions:
            if len(positions) == 1:
                newvalues.append(values[0])
            else:
                # current run:
                #       idxl pos idxr
                # p     p    |   p
                #    q       q
                idxl = max(0, bisect_left(positions, pos) - 1)
                idxr = min(len(positions) - 1, idxl + 1)
                fraction = (pos - positions[idxl]) / (positions[idxr] - positions[idxl])
                vall = values[idxl]
                valr = values[idxr]
                newval = func(vall, valr, fraction)
                newvalues.append(newval)
        return newvalues

    @staticmethod
    def append_to_doc(element, gradient):
        """Splits a gradient into stops and orientation, appends it to the document's
        defs and returns the href to the orientation gradient.

        Args:
            element (BaseElement): an element inside the SVG that the gradient should be
                added to
            gradient (Gradient): the gradient to append to the document

        Returns:
            Gradient: the orientation gradient, or the gradient object if
            element has no root or is None
        """
        stops, orientation = gradient.stops_and_orientation()
        if element is None or (
            element.getparent() is None and not isinstance(element, SvgDocumentElement)
        ):
            return gradient
        element.root.defs.add(orientation)
        if len(stops) > 0:
            element.root.defs.add(stops, orientation)
            orientation.set("xlink:href", f"#{stops.get_id()}")
        return orientation

    def interpolate(self, time=0):
        """Interpolate with another gradient."""
        newgrad = self.start_value.copy()
        # interpolate transforms
        newgrad.gradientTransform = self.transform_interpolator.interpolate(time)

        # interpolate orientation
        for attr in self.orientation_interpolator.keys():
            newgrad.set(attr, self.orientation_interpolator[attr].interpolate(time))

        # interpolate stops
        if self.newstop_interpolator is not None:
            newgrad.remove_all(Stop)
            newgrad.add(
                *[interp.interpolate(time) for interp in self.newstop_interpolator]
            )
        if self.svg is None:
            return newgrad
        return GradientInterpolator.append_to_doc(self.svg, newgrad)


class LinearGradientInterpolator(GradientInterpolator):
    """Class for interpolation of linear gradients"""

    def __init__(
        self, start_value=LinearGradient(), end_value=LinearGradient(), svg=None
    ):
        super().__init__(start_value, end_value, svg)

    @staticmethod
    def initialize_position(grad, bbox):
        """Initializes a linear gradient's position"""
        grad.set("x1", bbox.left)
        grad.set("x2", bbox.right)
        grad.set("y1", bbox.center.y)
        grad.set("y2", bbox.center.y)


class RadialGradientInterpolator(GradientInterpolator):
    """Class to interpolate radial gradients"""

    def __init__(
        self, start_value=RadialGradient(), end_value=RadialGradient(), svg=None
    ):
        super().__init__(start_value, end_value, svg)

    @staticmethod
    def initialize_position(grad, bbox):
        """Initializes a radial gradient's position"""
        x, y = bbox.center
        grad.set("cx", x)
        grad.set("cy", y)
        grad.set("fx", x)
        grad.set("fy", y)
        grad.set("r", bbox.right - bbox.center.x)


class StopInterpolator(AttributeInterpolator):
    """Class to interpolate gradient stops"""

    def __init__(self, start_value, end_value):
        super().__init__(start_value, end_value)
        self.style_interpolator = StyleInterpolator(start_value, end_value)
        self.position_interpolator = ValueInterpolator(
            float(start_value.offset), float(end_value.offset)
        )

    def interpolate(self, time=0):
        """Interpolates a gradient stop by interpolating style and offset separately

        Args:
            time (int, optional): Interpolation position. If 0, start_value is returned,
                if 1, end_value is returned. Defaults to 0.

        Returns:
            Stop: interpolated gradient stop
        """
        newstop = Stop()
        newstop.style = self.style_interpolator.interpolate(time)
        newstop.offset = self.position_interpolator.interpolate(time)
        return newstop


class PathInterpolator(AttributeInterpolator):
    """Base class for Path interpolation"""

    def __init__(self, start_value=Path(), end_value=Path()):
        super().__init__(start_value.to_superpath(), end_value.to_superpath())
        self.processed_end_path = None
        self.processed_start_path = None

    def truncate_subpaths(self):
        """Truncates the longer path so that all subpaths in both paths have an equal
        number of bezier commands"""
        s = [[]]
        e = [[]]
        # loop through all subpaths as long as there are remaining ones
        while self.start_value and self.end_value:
            # if both subpaths contain a bezier command, append it to s and e
            if self.start_value[0] and self.end_value[0]:
                s[-1].append(self.start_value[0].pop(0))
                e[-1].append(self.end_value[0].pop(0))
            # if the subpath of start_value is empty, add the remaining empty list as
            # new subpath of s and one more item of end_value as new subpath of e.
            # Afterwards, the loop terminates
            elif self.end_value[0]:
                s.append(self.start_value.pop(0))
                e[-1].append(self.end_value[0][0])
                e.append([self.end_value[0].pop(0)])
            elif self.start_value[0]:
                e.append(self.end_value.pop(0))
                s[-1].append(self.start_value[0][0])
                s.append([self.start_value[0].pop(0)])
            # if there are no commands left in both start_value or end_value, add empty
            # list to both start_value and end_value
            else:
                s.append(self.start_value.pop(0))
                e.append(self.end_value.pop(0))
        self.processed_start_path = s
        self.processed_end_path = e

    def interpolate(self, time=0):
        # create an interpolated path for each interval
        interp = []
        # process subpaths
        for ssubpath, esubpath in zip(
            self.processed_start_path, self.processed_end_path
        ):
            if not (ssubpath or esubpath):
                break
            # add a new subpath to the interpolated path
            interp.append([])
            # process each bezier command in the subpaths (which now have equal length)
            for sbezier, ebezier in zip(ssubpath, esubpath):
                if not (sbezier or ebezier):
                    break
                # add a new bezier command to the last subpath
                interp[-1].append([])
                # process points
                for point1, point2 in zip(sbezier, ebezier):
                    if not (point1 or point2):
                        break
                    # add a new point to the last bezier command
                    interp[-1][-1].append(
                        ArrayInterpolator(point1, point2).interpolate(time)
                    )
        # remove final subpath if empty.
        if not interp[-1]:
            del interp[-1]
        return CubicSuperPath(interp)


class EqualSubsegmentsInterpolator(PathInterpolator):
    """Interpolates the path by rediscretizing the subpaths first."""

    @staticmethod
    def get_subpath_lenghts(path):
        """prepare lengths for interpolation"""
        sp_lenghts, total = csplength(path)
        t = 0
        lenghts = []
        for sp in sp_lenghts:
            for l in sp:
                t += l / total
                lenghts.append(t)
        lenghts.sort()
        return sp_lenghts, total, lenghts

    @staticmethod
    def process_path(path, other):
        """Rediscretize path so that all subpaths have an equal number of segments,
        so that there is a node at the path "times" where path or other have a node

        Args:
            path (Path): the first path
            other (Path): the second path

        Returns:
            Array: the prepared path description for the intermediate path"""
        sp_lenghts, total, _ = EqualSubsegmentsInterpolator.get_subpath_lenghts(path)
        _, _, lenghts = EqualSubsegmentsInterpolator.get_subpath_lenghts(other)
        t = 0
        s = [[]]
        for sp in sp_lenghts:
            if not path[0]:
                s.append(path.pop(0))
            s[-1].append(path[0].pop(0))
            for l in sp:
                pt = t
                t += l / total
                if lenghts and t > lenghts[0]:
                    while lenghts and lenghts[0] < t:
                        nt = (lenghts[0] - pt) / (t - pt)
                        bezes = cspbezsplitatlength(s[-1][-1][:], path[0][0][:], nt)
                        s[-1][-1:] = bezes[:2]
                        path[0][0] = bezes[2]
                        pt = lenghts.pop(0)
                s[-1].append(path[0].pop(0))
        return s

    def __init__(self, start_path=Path(), end_path=Path()):
        super().__init__(start_path, end_path)
        # rediscretisize both paths
        start_copy = copy.deepcopy(self.start_value)
        # TODO find out why self.start_value.copy() doesn't work
        self.start_value = EqualSubsegmentsInterpolator.process_path(
            self.start_value, self.end_value
        )
        self.end_value = EqualSubsegmentsInterpolator.process_path(
            self.end_value, start_copy
        )

        self.truncate_subpaths()


class FirstNodesInterpolator(PathInterpolator):
    """Interpolates a path by discarding the trailing nodes of the longer subpath"""

    def __init__(self, start_path=Path(), end_path=Path()):
        super().__init__(start_path, end_path)
        # which path has fewer segments?
        lengthdiff = len(self.start_value) - len(self.end_value)
        # swap shortest first
        if lengthdiff > 0:
            self.start_value, self.end_value = self.end_value, self.start_value
        # subdivide the shorter path
        for _ in range(abs(lengthdiff)):
            maxlen = 0
            subpath = 0
            segment = 0
            for y, _ in enumerate(self.start_value):
                for z in range(1, len(self.start_value[y])):
                    leng = bezlenapprx(
                        self.start_value[y][z - 1], self.start_value[y][z]
                    )
                    if leng > maxlen:
                        maxlen = leng
                        subpath = y
                        segment = z
            sp1, sp2 = self.start_value[subpath][segment - 1 : segment + 1]
            self.start_value[subpath][segment - 1 : segment + 1] = cspbezsplit(sp1, sp2)
        # if swapped, swap them back
        if lengthdiff > 0:
            self.start_value, self.end_value = self.end_value, self.start_value
        self.truncate_subpaths()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2005,2007 Aaron Spike, aaron@ekips.org
# Copyright (C) 2009 Alvin Penner, penner@vaxxine.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension converts a path into a dashed line using 'stroke-dasharray'
It is a modification of the file addnodes.py
"""
import inkex
from inkex import bezier, CubicSuperPath, Group, PathElement
from inkex.localization import inkex_gettext as _


class Dashit(inkex.EffectExtension):
    """Extension to convert paths into dash-array line"""

    def __init__(self):
        super(Dashit, self).__init__()
        self.not_converted = []

    def effect(self):
        for node in self.svg.selection:
            self.convert2dash(node)
        if self.not_converted:
            inkex.errormsg(
                _("Total number of objects not converted: {}\n").format(
                    len(self.not_converted)
                )
            )
            # return list of IDs in case the user needs to find a specific object
            inkex.debug(self.not_converted)

    def convert2dash(self, node):
        """Convert each selected node's dash array"""
        if isinstance(node, Group):
            for child in node:
                self.convert2dash(child)
        elif isinstance(node, PathElement):
            self._convert(node)
        else:
            self.not_converted.append(node.get("id"))

    @staticmethod
    def _convert(node):
        dashes = []
        offset = 0
        style = node.specified_style()
        dashes = style("stroke-dasharray")
        offset = style("stroke-dashoffset")
        if not dashes:
            return
        new = []
        for sub in node.path.to_superpath():
            idash = 0
            dash = dashes[0]
            length = float(offset)
            while dash < length:
                length = length - dash
                idash = (idash + 1) % len(dashes)
                dash = dashes[idash]
            new.append([sub[0][:]])
            i = 1
            while i < len(sub):
                dash = dash - length
                length = bezier.cspseglength(new[-1][-1], sub[i])
                while dash < length:
                    new[-1][-1], nxt, sub[i] = bezier.cspbezsplitatlength(
                        new[-1][-1], sub[i], dash / length
                    )
                    if idash % 2:  # create a gap
                        new.append([nxt[:]])
                    else:  # splice the curve
                        new[-1].append(nxt[:])
                    length = length - dash
                    idash = (idash + 1) % len(dashes)
                    dash = dashes[idash]
                if idash % 2:
                    new.append([sub[i]])
                else:
                    new[-1].append(sub[i])
                i += 1
        style.pop("stroke-dasharray")
        node.pop("sodipodi:type")
        node.path = CubicSuperPath(new).to_path(rtol=1e-10)
        node.style = style


if __name__ == "__main__":
    Dashit().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2016 Richard White, rwhite8282@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
An Inkscape extension that creates a frame around a selected object.
"""

from typing import List

import inkex
from inkex import Group, PathElement, ClipPath
from inkex.localization import inkex_gettext as _


class Frame(inkex.EffectExtension):
    """
    An Inkscape extension that creates a frame around a selected object.
    """

    def add_arguments(self, pars):
        # Parse the options.
        pars.add_argument("--tab", default="stroke")
        pars.add_argument("--clip", type=inkex.Boolean, default=False)
        pars.add_argument("--type", default="rect", choices=["rect", "ellipse"])
        pars.add_argument("--corner_radius", type=int, default=0)
        pars.add_argument("--fill_color", type=inkex.Color, default=inkex.Color(0))
        pars.add_argument("--group", type=inkex.Boolean, default=False)
        pars.add_argument("--stroke_color", type=inkex.Color, default=inkex.Color(0))
        pars.add_argument("--width", type=float, default=2.0)
        pars.add_argument(
            "--offset_absolute",
            type=float,
            default=0,
            help="Offset in user units, positive = outside",
        )
        pars.add_argument(
            "--offset_relative",
            type=float,
            default=0,
            help="Relative offset in percentage of the bounding box size",
        )
        pars.add_argument(
            "--z_position",
            type=str,
            default="bottom",
            choices=["top", "bottom", "split"],
        )
        pars.add_argument("--asgroup", type=inkex.Boolean, default=False)

    def add_clip(self, node: inkex.BaseElement, clip_path: inkex.PathElement):
        """Adds a new clip path node to the defs and sets
            the clip-path on the node.
        node -- The node that will be clipped.
        clip_path -- The clip path object.
        """
        clip = ClipPath()
        # apply the reverse transform to the clip path. no composed_transform here,
        # since the frame will be appended to the object' parent (or an untransformed
        # group within).
        clip.append(PathElement.new(path=clip_path.path, transform=-node.transform))
        self.svg.defs.add(clip)
        node.set("clip-path", clip.get_id(as_url=2))

    @staticmethod
    def generate_frame(name, box: inkex.BoundingBox, rectangle=True, radius=0):
        """
        name -- The name of the new frame object.
        box -- The boundary box of the node.
        style -- The style used to draw the path.
        radius -- The corner radius of the frame.
        returns a new frame node.
        """
        if rectangle:
            r = min([radius, abs(box.x.size / 2), abs(box.y.size / 2)])
            elem = inkex.Rectangle.new(
                left=box.x.minimum,
                top=box.y.minimum,
                width=box.x.size,
                height=box.y.size,
                rx=r,
            )
        else:
            elem = inkex.Ellipse.new(center=box.center, radius=box.size / 2)
        elem.label = name
        return elem

    def create_frame(self, containedelements: List[inkex.BaseElement]):
        """generate the frame for an element or a group of elements"""
        width = self.options.width
        style = inkex.Style({"stroke-width": width})
        style.set_color(self.options.stroke_color, "stroke")
        elem_top = None
        elem_bottom = None

        box = inkex.BoundingBox()
        for node in containedelements:
            if isinstance(node, (inkex.TextElement, inkex.Tspan, inkex.FlowRoot)):
                try:
                    box += node.get_inkscape_bbox()
                except ValueError:
                    continue
            else:
                box += node.bounding_box()
        box = box.resize(
            box.x.size * (self.options.offset_relative / 100),
            box.y.size * (self.options.offset_relative / 100),
        )
        box = box.resize(self.options.offset_absolute)

        frame = self.generate_frame(
            "Frame", box, self.options.type == "rect", self.options.corner_radius
        )
        if self.options.z_position != "split":
            style.set_color(self.options.fill_color, "fill")
            frame.style = style
            if self.options.z_position == "bottom":
                elem_bottom = frame
            else:
                elem_top = frame
        else:
            fill = frame.copy()
            elem_top = frame
            elem_top.style = style
            elem_bottom = fill
            elem_bottom.style.set_color(self.options.fill_color, "fill")
            elem_top.style["fill"] = None
        return elem_bottom, elem_top

    def process_elements(self, containedelements: List[inkex.BaseElement]):
        """Create and append the frame for an object or a set of objects."""
        elem_bottom, elem_top = self.create_frame(containedelements)
        if self.options.clip:
            for node in containedelements:
                element = elem_top if elem_top is not None else elem_bottom
                self.add_clip(node, element)
        if self.options.group:
            group = containedelements[0].getparent().add(Group())
            if elem_bottom is not None:
                group.append(elem_bottom)
            for node in containedelements:
                group.append(node)
            if elem_top is not None:
                group.append(elem_top)
        else:
            if elem_bottom is not None:
                containedelements[0].addprevious(elem_bottom)
            if elem_top is not None:
                containedelements[-1].addnext(elem_top)

    def effect(self):
        """Performs the effect."""
        # Determine common properties.
        if not self.svg.selection:
            raise inkex.AbortExtension(_("Select at least one object."))
        style = inkex.Style({"stroke-width": self.options.width})
        style.set_color(self.options.stroke_color, "stroke")

        if not self.options.asgroup:

            for node in self.svg.selection:
                self.process_elements([node])
        else:
            self.process_elements(self.svg.selection.rendering_order())


if __name__ == "__main__":
    Frame().run()



#!/usr/bin/env python
"""
Convert PNG to Jpeg using Raster Output extension.
"""

import inkex


class JpegOutput(inkex.RasterOutputExtension):
    multi_inx = True  # XXX Remove this after refactoring

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--quality", type=int, default=90)
        pars.add_argument("--progressive", type=inkex.Boolean, default=False)

    def save(self, stream):
        self.img.convert("RGB").save(
            stream,
            format="jpeg",
            quality=self.options.quality,
            progressive=self.options.progressive,
        )


if __name__ == "__main__":
    JpegOutput().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2008 Stephen Silver
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
"""
Simple wrapper around fig2dev
"""

import inkex
from inkex.command import call


class FigInput(inkex.CallExtension):
    """Load FIG Files by calling fig2dev program"""

    input_ext = "fig"

    def call(self, input_file, output_file):
        call("fig2dev", "-L", "svg", input_file, output_file)


if __name__ == "__main__":
    FigInput().run()



#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sys
import json
import logging

from import_sources import RemoteSource, RemoteFile
from urllib.parse import urljoin, parse_qs

try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None


class OpenClipartFile(RemoteFile):
    def get_file(self):
        """Extract search result from html"""
        response = self.remote.session.get(self.info["file"])
        soup = BeautifulSoup(response.text, features="lxml")
        for script in soup.find_all("script"):
            content = script.contents
            if content and "image" in content[0]:
                try:
                    data = json.loads(content[0])
                    return self.remote.to_local_file(data["image"]["url"])
                except Exception:
                    continue
        logging.error("Couldn't load svg from %s", self.info["file"])


class OpenClipart(RemoteSource):
    name = "Open Clipart Library"
    icon = "sources/ocal.svg"
    base_url = "https://openclipart.org/search/"
    is_enabled = BeautifulSoup is not None
    file_cls = OpenClipartFile

    def html_search(self, response):
        """Extract search results from html"""
        soup = BeautifulSoup(response.text, features="lxml")
        for div in soup.find_all("div", {"class": "artwork"}):
            if div.a and div.a.img:
                link = urljoin(self.base_url, div.a.get("href"))
                img = urljoin(self.base_url, div.a.img.get("src"))

                yield {
                    "file": link,  # Not the actual file yet (see above)
                    "name": div.a.img.get("alt"),
                    "thumbnail": img,
                    "author": "OpenClipart",
                    "license": "cc-0",
                }

        for page in soup.find_all("a", {"class": "page-link", "aria-label": "Next"}):
            if "=" in page.get("href", ""):
                yield lambda: self._search(**parse_qs(page.get("href").split("?")[-1]))

    def search(self, query):
        """HTML searching for now"""
        return self._search(query=query)

    def _search(self, **params):
        try:
            response = self.session.get(self.base_url, params=params)
        except Exception:
            return []

        items = []
        next_page = None
        for item in self.html_search(response):
            if callable(item):
                next_page = item
            else:
                items.append(item)
        # Often ocal will have empty pages, weirdly.
        if not items and next_page:
            return next_page()
        # None empty page, return all
        if next_page:
            items.append(next_page)
        return items



# coding=utf-8
import os
from string import Template
import xml.etree.ElementTree as ET
import argparse

parser = argparse.ArgumentParser(
    description="Reads an *.inx file and generates initialization code for argparse"
)
parser.add_argument("input")
args = parser.parse_args()

if os.path.isabs(args.input):
    inpath = args.input
else:
    folder = os.path.dirname(os.path.realpath(__file__))
    inpath = os.path.normpath(os.path.join(folder, args.input))

templateWithType = Template(
    'self.arg_parser.add_argument("--$param",  type=$type, dest="$param", default=$default)'
)
templateWithoutType = Template(
    'self.arg_parser.add_argument("--$param",  dest="$param", default=$default)'
)


def handle_param_node(node):
    if node.attrib["type"] == "float":
        cmd = templateWithType.substitute(
            param=node.attrib["name"], type="float", default=node.text
        )
        print(cmd)
    elif node.attrib["type"] == "int":
        cmd = templateWithType.substitute(
            param=node.attrib["name"], type="int", default=node.text
        )
        print(cmd)
    elif node.attrib["type"] == "boolean":
        cmd = templateWithType.substitute(
            param=node.attrib["name"],
            type="inkex.inkbool",
            default='"' + node.text + '"',
        )
        print(cmd)
    elif node.attrib["type"] == "enum":
        cmd = templateWithoutType.substitute(
            param=node.attrib["name"], default='"' + node[0].text + '"'
        )
        print(cmd)
    elif node.attrib["type"] == "notebook":
        cmd = templateWithoutType.substitute(
            param=node.attrib["name"], default='"' + node[0].attrib["name"] + '"'
        )
        print(cmd)
    else:
        # TODO: Implement other types of args
        raise NotImplementedError


def process_node(node):
    for child in node:
        if child.tag.endswith("param"):
            handle_param_node(child)
        process_node(child)


tree = ET.parse(inpath)
root = tree.getroot()
process_node(root)



# coding=utf-8
#
# Copyright (C) 2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Property management and parsing, CSS cascading, default value storage

.. versionadded:: 1.2

.. data:: all_properties

    A list of all properties, their parser class, and additional information
    such as whether they are inheritable or can be given as presentation attributes
"""

from abc import ABC, abstractmethod

import re
from typing import Tuple, Dict, Type, Union, List, Optional
from .interfaces.IElement import IBaseElement, ISVGDocumentElement

from .units import parse_unit, convert_unit

from .colors import Color, ColorError


class BaseStyleValue:
    """A class encapsuling a single CSS declaration / presentation_attribute"""

    def __init__(self, declaration=None, attr_name=None, value=None, important=False):
        self.attr_name: str
        self.value: str
        self.important: bool
        if declaration is not None and ":" in declaration:
            (
                self.attr_name,
                self.value,
                self.important,
            ) = BaseStyleValue.parse_declaration(declaration)
        elif attr_name is not None:
            self.attr_name = attr_name.strip().lower()
            if isinstance(value, str):
                self.value = value.strip()
            else:
                # maybe its already parsed? then set it
                self.value = self.unparse_value(value)
            self.important = important
            _ = self.parse_value()  # check that we can parse this value

    @classmethod
    def parse_declaration(cls, declaration: str) -> Tuple[str, str, bool]:
        """Parse a single css declaration

        Args:
            declaration (str): a css declaration such as:
                ``fill: #000 !important;``. The trailing semicolon may be ommitted.

        Raises:
            ValueError: Unable to parse the declaration

        Returns:
            Tuple[str, str, bool]: a tuple with key, value and importance
        """
        if declaration != "" and ":" in declaration:
            declaration = declaration.replace(";", "")
            (name, value) = declaration.split(":", 1)
            # check whether this is an important declaration
            important = False
            if "!important" in value:
                value = value.replace("!important", "")
                important = True
            return (name.strip().lower(), value.strip(), important)
        raise ValueError("Invalid declaration")

    def parse_value(self, element=None):
        """Get parsed property value with resolved urls, color, etc.

        Args:
            element (BaseElement): the SVG element to which this style is applied to
                currently used for resolving gradients / masks, could be used for
                computing percentage attributes or calc() attributes [optional]

        Returns:
            object: parsed property value
        """
        if self.value == "inherit":
            if self.attr_name in all_properties:
                return self._parse_value(all_properties[self.attr_name][1], element)
            return None
        return self._parse_value(self.value, element)

    def _parse_value(  # pylint: disable=unused-argument, no-self-use
        self, value: str, element=None
    ) -> object:
        """internal parse method, to be overwritten by derived classes

        Args:
            value (str): unparsed value
            element (BaseElement): the SVG element to which this style is applied to
                [optional]

        Returns:
            object: the parsed value
        """
        return value

    def unparse_value(self, value: object) -> str:
        """ "Unparses" an object, i.e. converts an object back to an attribute.
        If the result is invalid, i.e. can't be parsed, an exception is raised.

        Args:
            value (object): the object to be unparsed

        Returns:
            str: the attribute value
        """
        result = self._unparse_value(value)
        self._parse_value(result)  # check if value can be parsed (value is valid)
        return result

    def _unparse_value(self, value: object) -> str:  # pylint: disable=no-self-use
        return str(value)

    @property
    def declaration(self) -> str:
        """The css declaration corresponding to the StyleValue object

        Returns:
            str: the css declaration, such as "fill: #000 !important;"
        """
        return (
            self.attr_name
            + ":"
            + self.value
            + (" !important" if self.important else "")
        )

    @classmethod
    def factory(
        cls,
        declaration: Optional[str] = None,
        attr_name: Optional[str] = None,
        value: Optional[object] = None,
        important: Optional[bool] = False,
    ):
        """Create an attribute

        Args:
            declaration (str, optional): the CSS declaration to parse. Defaults to None.
            attr_name (str, optional): the attribute name. Defaults to None.
            value (object, optional): the attribute value. Defaults to None.
            important (bool, optional): whether the attribute is marked !important.
                Defaults to False.

        Raises:
            Errors may also be raised on parsing, so make sure to handle them

        Returns:
            BaseStyleValue: the parsed style
        """
        if declaration is not None and ":" in declaration:
            attr_name, value, important = BaseStyleValue.parse_declaration(declaration)
        elif attr_name is not None and value is not None:
            attr_name = attr_name.strip().lower()
            if isinstance(value, str):
                value = value.strip()

        if attr_name in all_properties:
            valuetype = all_properties[attr_name][0]
            return valuetype(declaration, attr_name, value, important)
        return BaseStyleValue(declaration, attr_name, value, important)

    def __eq__(self, other):
        if not (isinstance(other, BaseStyleValue)):
            return False
        if self.declaration != other.declaration:
            return False
        return True

    @staticmethod
    def factory_errorhandled(element=None, declaration="", key="", value=""):
        """Error handling for the factory method: if something goes wrong during
        parsing, ignore the attribute

        Args:
            element (BaseElement, optional): The element this declaration is affecting,
                for finding gradients ect. Defaults to None.
            declaration (str, optional): the CSS declaration to parse. Defaults to "".
            key (str, optional): the attribute name. Defaults to "".
            value (str, optional): the attribute value. Defaults to "".

        Returns:
            BaseStyleValue: The parsed style
        """
        try:
            value = BaseStyleValue.factory(
                declaration=declaration, attr_name=key, value=value
            )
            key = value.attr_name
            # Try to parse the attribute
            _ = value.parse_value(element)
            return (key, value)
        except ValueError:
            # something went wrong during parsing, e.g. a bad attribute format
            # or an attribute referencing a missing gradient
            # -> ignore this declaration
            pass
        except ColorError:
            # The color parsing methods have their own error type
            pass


class AlphaValue(BaseStyleValue):
    """Stores an alpha value (such as opacity), which may be specified as
    as percentage or absolute value.

    Reference: https://www.w3.org/TR/css-color/#typedef-alpha-value"""

    def _parse_value(self, value: str, element=None):
        if value[-1] == "%":  # percentage
            parsed_value = float(value[:-1]) * 0.01
        else:
            parsed_value = float(value)
        if parsed_value < 0:
            return 0
        if parsed_value > 1:
            return 1
        return parsed_value

    def _unparse_value(self, value: object) -> str:
        if isinstance(value, (float, int)):
            if value < 0:
                return "0"
            if value > 1:
                return "1"
            return str(value)
        raise ValueError("Value must be number")


class ColorValue(BaseStyleValue):
    """Stores a color value

    Reference: https://drafts.csswg.org/css-color-3/#valuea-def-color"""

    def _parse_value(self, value: str, element=None):
        if value == "currentColor":
            if element is not None:
                style = element.specified_style()
                return style("color")
            return None
        return Color(value)


# https://www.w3.org/TR/css3-values/#url-value
# matches anything inside url() enclosed with single/double quotes
# (technically a fragment url) or no quotes at all
URLREGEX = r"url\(\s*('.*?'|\".*?\"|[^\"'].*?)\s*\)"


def match_url_and_return_element(string: str, svg):
    """Parses a string containing an url, e.g. "url(#rect1234)",
    looks up the element in the svg document and returns it.

    Args:
        string (str): the string to parse
        svg (SvgDocumentElement): document referenced in the URL

    Raises:
        ValueError: if the string has invalid format
        ValueError: if the referenced element is not found

    Returns:
        BaseElement: the referenced element
    """
    regex = re.compile(URLREGEX)
    match = regex.match(string)
    if match:
        url = match.group(1)
        paint_server = svg.getElementById(url)
        return paint_server
    raise ValueError("invalid URL format")


class URLNoneValue(BaseStyleValue):
    """Stores a value that is either none or an url, such as markers or masks.

    Reference: https://www.w3.org/TR/SVG2/painting.html#VertexMarkerProperties"""

    def _parse_value(self, value: str, element=None):
        if value == "none":
            return None
        if value[0:4] == "url(":
            if element is not None and self.element_has_root(element):
                return match_url_and_return_element(value, element.root)
            return None
        raise ValueError("Invalid property value")

    def _unparse_value(self, value: object):
        if isinstance(value, IBaseElement):
            return f"url(#{value.get_id()})"
        return super()._unparse_value(value)

    @staticmethod
    def element_has_root(element) -> bool:
        "Checks if an element has a root, i.e. if element.root will fail"

        return not (
            element.getparent() is None and not isinstance(element, ISVGDocumentElement)
        )


class PaintValue(ColorValue, URLNoneValue):
    """Stores a paint value (such as fill and stroke), which may be specified
    as color, or url.

    Reference: https://www.w3.org/TR/SVG2/painting.html#SpecifyingPaint"""

    def _parse_value(self, value: str, element=None):
        if value == "none":
            return None
        if value in ["context-fill", "context-stroke"]:
            return value
        if value == "currentColor":
            return super()._parse_value(value, element)
        if value[0:4] == "url(":
            # First part: fragment url
            # second part: a fallback color if the url is not found. Colors start with
            # a letter or a #
            regex = re.compile(URLREGEX + r"\s*([#\w].*?)?$")
            match = regex.match(value)
            if match:
                url = match.group(1)
                if element is not None and self.element_has_root(element):
                    paint_server = element.root.getElementById(url)
                else:
                    return None
                if paint_server is not None:
                    return paint_server
                if match.group(2) is None:
                    raise ValueError("Paint server not found")
                return Color(match.group(2))
        return Color(value)

    def _unparse_value(self, value: object):
        if value is None:
            return "none"
        return super()._unparse_value(value)


class EnumValue(BaseStyleValue):
    """Stores a value that can only have a finite set of options"""

    def __init__(self, declaration=None, attr_name=None, value=None, important=False):
        self.valueset = all_properties[attr_name][4]
        super().__init__(declaration, attr_name, value, important)

    def _parse_value(self, value: str, element=None):
        if value in self.valueset:
            return value
        raise ValueError(
            f"Value '{value}' is invalid for the property {self.attr_name}. "
            + f"Allowed values are: {self.valueset + ['inherit']}"
        )


class ShorthandValue(BaseStyleValue, ABC):
    """Stores a value that sets other values (e.g. the font shorthand)"""

    def apply_shorthand(self, style):
        """Applies a shorthand attribute to its style, respecting the
        importance state of the individual attributes.

        Args:
            style (Style): the style that the shorthand attribute is contained in,
                and that the shorthand attribute will be applied on
        """
        if self.attr_name not in style:
            return

        dct = self.get_shorthand_changes()
        importance = self.important

        # they are ordered in the order of adding the style elements.
        current_keys = list(style.keys())
        for curkey in dct:
            perform = False
            if curkey not in current_keys:
                # this is the easiest case, just set the element and the importance
                perform = True
            else:
                if importance != style.get_importance(curkey):
                    # different importance, result independent of position
                    perform = importance
                else:
                    # only apply if style overwrites previous with same importance
                    perform = current_keys.index(curkey) < current_keys.index(
                        self.attr_name
                    )

            if perform:
                style[curkey] = dct[curkey]
                style.set_importance(curkey, importance)
        style.pop(self.attr_name)

    @abstractmethod
    def get_shorthand_changes(self) -> Dict[str, str]:
        """calculates the value of affected attributes for this shorthand

        Returns:
            Dict[str, str]: a dictionary containing the new values of the
            affected attributes
        """


class FontValue(ShorthandValue):
    """Logic for the shorthand font property"""

    def get_shorthand_changes(self):
        keys = [
            "font-style",
            "font-variant",
            "font-weight",
            "font-stretch",
            "font-size",
            "line-height",
            "font-family",
        ]
        options = {
            key: all_properties[key][4]
            for key in keys
            if isinstance(all_properties[key][4], list)
        }
        result = {key: all_properties[key][1] for key in keys}

        tokens = [i for i in self.value.split(" ") if i != ""]

        if len(tokens) == 0:
            return {}  # shorthand not set, nothing to do

        while not (len(tokens) == 0):
            cur = tokens[0]
            if cur in options["font-style"]:
                result["font-style"] = cur
            elif cur in options["font-variant"]:
                result["font-variant"] = cur
            elif cur in options["font-weight"]:
                result["font-weight"] = cur
            elif cur in options["font-stretch"]:
                result["font-stretch"] = cur
            else:
                if "/" in cur:
                    result["font-size"], result["line-height"] = cur.split("/")
                else:
                    result["font-size"] = cur
                result["font-family"] = " ".join(tokens[1:])
                break
            tokens = tokens[1:]  # remove first element
        return result


class MarkerShorthandValue(ShorthandValue, URLNoneValue):
    """Logic for the marker shorthand property"""

    def get_shorthand_changes(self):
        if self.value == "":
            return {}  # shorthand not set, nothing to do
        return {k: self.value for k in ["marker-start", "marker-end", "marker-mid"]}

    def _parse_value(self, value: str, element=None):
        # Make sure the parsing routine doesn't choke on an empty shorthand
        if value == "":
            return ""
        return super()._parse_value(value, element)


class FontSizeValue(BaseStyleValue):
    """Logic for the font-size property"""

    def _parse_value(self, value: str, element=None):
        if element is None:
            return value  # no additional logic in this case
        try:
            return element.to_dimensionless(value)
        except ValueError:  # unable to parse font size, e.g. font-size:normal
            return element.to_dimensionless("12pt")


class StrokeDasharrayValue(BaseStyleValue):
    """Logic for the stroke-dasharray property"""

    def _parse_value(self, value: str, element=None):
        dashes = re.findall(r"[^,\s]+", value)
        if len(dashes) == 0 or value == "none":
            return None  # no dasharray applied
        if not any(parse_unit(i) is None for i in dashes):
            dashes = [convert_unit(i, "px") for i in dashes]
        else:
            return None
        if any(i < 0 for i in dashes):
            return None  # one negative value makes the dasharray invalid
        if len(dashes) % 2 == 1:
            dashes = 2 * dashes
        return dashes

    def _unparse_value(self, value: object) -> str:
        if value is None:
            return "none"
        if isinstance(value, list):
            return " ".join(map(str, value))
        return str(value)


# keys: attributes, right side:
# - Subclass of BaseStyleValue used for instantiating
# - default value
# - is presentation attribute
# - inherited
# - additional information, such as valid enum values
# For properties which have no special implementation yet:
# "(BaseStyleValue, <default>, <inheritance>, None)"

# Source for this list: https://www.w3.org/TR/SVG2/styling.html#PresentationAttributes


all_properties: Dict[
    str, Tuple[Type[BaseStyleValue], str, bool, bool, Union[List[str], None]]
] = {
    "alignment-baseline": (
        EnumValue,
        "baseline",
        True,
        False,
        [
            "baseline",
            "text-bottom",
            "alphabetic",
            "ideographic",
            "middle",
            "central",
            "mathematical",
            "text-top",
        ],
    ),
    "baseline-shift": (BaseStyleValue, "0", True, False, None),
    "clip": (BaseStyleValue, "auto", True, False, None),
    "clip-path": (URLNoneValue, "none", True, False, None),
    "clip-rule": (EnumValue, "nonzero", True, True, ["nonzero", "evenodd"]),
    # only used for currentColor, which is not yet implemented
    "color": (PaintValue, "black", True, True, None),
    "color-interpolation": (
        EnumValue,
        "sRGB",
        True,
        True,
        ["sRGB", "auto", "linearRGB"],
    ),
    "color-interpolation-filters": (
        EnumValue,
        "linearRGB",
        True,
        True,
        ["auto", "sRGB", "linearRGB"],
    ),
    "color-rendering": (
        EnumValue,
        "auto",
        True,
        True,
        ["auto", "optimizeSpeed", "optimizeQuality"],
    ),
    "cursor": (BaseStyleValue, "auto", True, True, None),
    "direction": (EnumValue, "ltr", True, True, ["ltr", "rtl"]),
    "display": (
        EnumValue,
        "inline",
        True,
        False,
        [
            "inline",
            "block",
            "list-item",
            "inline-block",
            "table",
            "inline-table",
            "table-row-group",
            "table-header-group",
            "table-footer-group",
            "table-row",
            "table-column-group",
            "table-column",
            "table-cell",
            "table-caption",
            "none",
        ],
    ),  # every value except none is rendered normally
    "dominant-baseline": (
        EnumValue,
        "auto",
        True,
        True,
        [
            "auto",
            "text-bottom",
            "alphabetic",
            "ideographic",
            "middle",
            "central",
            "mathematical",
            "hanging",
            "text-top",
        ],
    ),
    "fill": (
        PaintValue,
        "black",
        True,
        True,
        None,
    ),  # the normal fill, not the <animation> one
    "fill-opacity": (AlphaValue, "1", True, True, None),
    "fill-rule": (EnumValue, "nonzero", True, True, ["nonzero", "evenodd"]),
    "filter": (BaseStyleValue, "none", True, False, None),
    "flood-color": (PaintValue, "black", True, False, None),
    "flood-opacity": (AlphaValue, "1", True, False, None),
    "font": (FontValue, "", True, False, None),
    "font-family": (BaseStyleValue, "sans-serif", True, True, None),
    "font-size": (FontSizeValue, "medium", True, True, None),
    "font-size-adjust": (BaseStyleValue, "none", True, True, None),
    "font-stretch": (
        EnumValue,
        "normal",
        True,
        True,
        [
            "normal",
            "ultra-condensed",
            "extra-condensed",
            "condensed",
            "semi-condensed",
            "semi-expanded",
            "expanded",
            "extra-expanded",
            "ultra-expanded",
        ],
    ),
    "font-style": (EnumValue, "normal", True, True, ["normal", "italic", "oblique"]),
    # a lot more values and subproperties in SVG2 / CSS-Fonts3
    "font-variant": (EnumValue, "normal", True, True, ["normal", "small-caps"]),
    "font-weight": (
        EnumValue,
        "normal",
        True,
        True,
        ["normal", "bold"] + [str(i) for i in range(100, 901, 100)],
    ),
    "glyph-orientation-horizontal": (BaseStyleValue, "0deg", True, True, None),
    "glyph-orientation-vertical": (BaseStyleValue, "auto", True, True, None),
    "image-rendering": (
        EnumValue,
        "auto",
        True,
        True,
        ["auto", "optimizeQuality", "optimizeSpeed"],
    ),
    "letter-spacing": (BaseStyleValue, "normal", True, True, None),
    "lighting-color": (ColorValue, "normal", True, False, None),
    "line-height": (BaseStyleValue, "normal", False, True, None),
    "marker": (MarkerShorthandValue, "", True, True, None),
    "marker-end": (URLNoneValue, "none", True, True, None),
    "marker-mid": (URLNoneValue, "none", True, True, None),
    "marker-start": (URLNoneValue, "none", True, True, None),
    # is a shorthand for a lot of mask-related properties which Inkscape doesn't support
    "mask": (URLNoneValue, "none", True, False, None),
    "opacity": (AlphaValue, "1", True, False, None),
    "overflow": (
        EnumValue,
        "visible",
        True,
        False,
        ["visible", "hidden", "scroll", "auto"],
    ),
    "paint-order": (BaseStyleValue, "normal", True, False, None),
    "pointer-events": (
        EnumValue,
        "visiblePainted",
        True,
        True,
        [
            "bounding-box",
            "visiblePainted",
            "visibleFill",
            "visibleStroke",
            "visible",
            "painted",
            "fill",
            "stroke",
            "all",
            "none",
        ],
    ),
    "shape-rendering": (
        EnumValue,
        "visiblePainted",
        True,
        True,
        ["auto", "optimizeSpeed", "crispEdges", "geometricPrecision"],
    ),
    "stop-color": (ColorValue, "black", True, False, None),
    "stop-opacity": (AlphaValue, "1", True, False, None),
    "stroke": (PaintValue, "none", True, True, None),
    "stroke-dasharray": (StrokeDasharrayValue, "none", True, True, None),
    "stroke-dashoffset": (BaseStyleValue, "0", True, True, None),
    "stroke-linecap": (EnumValue, "butt", True, True, ["butt", "round", "square"]),
    "stroke-linejoin": (
        EnumValue,
        "miter",
        True,
        True,
        ["miter", "miter-clip", "round", "bevel", "arcs"],
    ),
    "stroke-miterlimit": (BaseStyleValue, "4", True, True, None),
    "stroke-opacity": (AlphaValue, "1", True, True, None),
    "stroke-width": (BaseStyleValue, "1", True, True, None),
    "text-align": (
        BaseStyleValue,
        "start",
        True,
        True,
        None,
    ),  # only HTML property, but used by some unit tests
    "text-anchor": (EnumValue, "start", True, True, ["start", "middle", "end"]),
    # shorthand for text-decoration-line, *-style, *-color
    "text-decoration": (BaseStyleValue, "none", True, True, None),
    "text-overflow": (EnumValue, "clip", True, False, ["clip", "ellipsis"]),
    "text-rendering": (
        EnumValue,
        "auto",
        True,
        True,
        ["auto", "optimizeSpeed", "optimizeLegibility", "geometricPrecision"],
    ),
    "unicode-bidi": (
        EnumValue,
        "normal",
        True,
        False,
        [
            "normal",
            "embed",
            "isolate",
            "bidi-override",
            "isolate-override",
            "plaintext",
        ],
    ),
    "vector-effect": (BaseStyleValue, "none", True, False, None),
    "vertical-align": (BaseStyleValue, "baseline", False, False, None),
    "visibility": (EnumValue, "visible", True, True, ["visible", "hidden", "collapse"]),
    "white-space": (
        EnumValue,
        "normal",
        True,
        True,
        ["normal", "pre", "nowrap", "pre-wrap", "break-spaces", "pre-line"],
    ),
    "word-spacing": (BaseStyleValue, "normal", True, True, None),
    # including obsolete SVG 1.1 values
    "writing-mode": (
        EnumValue,
        "horizontal-tb",
        True,
        True,
        [
            "horizontal-tb",
            "vertical-rl",
            "vertical-lr",
            "lr",
            "lr-tb",
            "rl",
            "rl-tb",
            "tb",
            "tb-rl",
        ],
    ),
    "-inkscape-font-specification": (BaseStyleValue, "sans-serif", False, True, None),
}



# coding=utf-8
#
# Copyright (C) 2014 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=no-self-use
"""
Renderer for barcodes, SVG extension for Inkscape.

For supported barcodes see Barcode module directory.
"""

from inkex.localization import inkex_gettext as _

# This lists all known Barcodes missing from this package
# ===== UPC-Based Extensions ====== #
# Code11
# ========= Code25-Based ========== #
# Codabar
# Postnet
# ITF25
# ========= Alpha-numeric ========= #
# Code39Mod
# USPS128
# =========== 2D Based ============ #
# PDF417
# PDF417-Macro
# PDF417-Truncated
# PDF417-GLI


class NoBarcode:
    """Simple class for no barcode"""

    def __init__(self, msg):
        self.msg = msg

    def encode(self, text):
        """Encode the text into a barcode pattern"""
        raise ValueError(_("No barcode encoder: {}").format(self.msg))

    def generate(self):
        """Generate actual svg from the barcode pattern"""
        return None


def get_barcode(code, **kw):
    """Gets a barcode from a list of available barcode formats"""
    if not code:
        return NoBarcode(_("No barcode format given."))

    code = str(code).replace("-", "").strip()
    module = "barcode." + code
    lst = ["barcode"]
    try:
        return getattr(__import__(module, fromlist=lst), code)(kw)
    except ImportError as err:
        if code in str(err):
            return NoBarcode(_("Invalid type of barcode: {}.{}").format(module, code))
        raise
    except AttributeError:
        return NoBarcode(
            _("Barcode module is missing barcode class: {}.{}").format(module, code)
        )



#
# Copyright (c) Gary Wilson Jr. <gary@thegarywilson.com> and contributors.
#
# MIT License, see https://github.com/gdub/python-archive/blob/master/LICENSE.txt
#
"""
Compatability for zip and tar files, taken from upstream and stripped down
to provide list and read interface, removed PY2 support.
"""

import os
import sys
import tarfile
import zipfile


class ArchiveException(Exception):
    """Base exception class for all archive errors."""


class UnrecognizedArchiveFormat(ArchiveException):
    """Error raised when passed file is not a recognized archive format."""


class Archive(object):
    """
    The external API class that encapsulates an archive implementation.
    """

    def __init__(self, filename, ext=""):
        """
        Arguments:
        * 'file' can be a string path to a file or a file-like object.
        * Optional 'ext' argument can be given to override the file-type
          guess that is normally performed using the file extension of the
          given 'file'.  Should start with a dot, e.g. '.tar.gz'.
        """
        cls = self._archive_cls(filename, ext=ext)
        self._archive = cls(filename)

    def __enter__(self):
        return self._archive

    def __exit__(self, exc, value, traceback):
        pass

    @staticmethod
    def _archive_cls(filename, ext=""):
        """
        Return the proper Archive implementation class, based on the file type.
        """
        cls = None
        if not isinstance(filename, str):
            try:
                filename = filename.name
            except AttributeError:
                raise UnrecognizedArchiveFormat(
                    "File object not a recognized archive format."
                )
        lookup_filename = filename + ext
        base, tail_ext = os.path.splitext(lookup_filename.lower())
        cls = extension_map.get(tail_ext)
        if not cls:
            base, ext = os.path.splitext(base)
            cls = extension_map.get(ext)
        if not cls:
            raise UnrecognizedArchiveFormat(
                "Path not a recognized archive format: %s" % filename
            )
        return cls


class BaseArchive(object):
    def __del__(self):
        if hasattr(self, "_archive"):
            self._archive.close()

    def list(self):
        raise NotImplementedError()

    def filenames(self):
        raise NotImplementedError()


class TarArchive(BaseArchive):
    def __init__(self, filename):
        # tarfile's open uses different parameters for file path vs. file obj.
        if isinstance(filename, str):
            self._archive = tarfile.open(name=filename)
        else:
            self._archive = tarfile.open(fileobj=filename)

    def filenames(self):
        return self._archive.getnames()

    def read(self, name):
        return self._archive.extractfile(name).read()


class ZipArchive(BaseArchive):
    def __init__(self, file):
        # ZipFile's 'file' parameter can be path (string) or file-like obj.
        self._archive = zipfile.ZipFile(file)

    def filenames(self):
        return self._archive.namelist()

    def read(self, name):
        return self._archive.read(name)


extension_map = {
    ".egg": ZipArchive,
    ".zip": ZipArchive,
    ".tar": TarArchive,
    ".tar.bz2": TarArchive,
    ".tar.gz": TarArchive,
    ".tgz": TarArchive,
    ".tz2": TarArchive,
}



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2011 Felipe Correa da Silva Sanches
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import locale
import sys

import inkex


class NewGlyphLayer(inkex.EffectExtension):
    def add_arguments(self, pars):
        self.arg_parser.add_argument("--text", default="", help="Unicode chars")

        self.encoding = sys.stdin.encoding
        if self.encoding == "cp0" or self.encoding is None:
            self.encoding = locale.getpreferredencoding()

    def effect(self):
        # Get all the options
        unicode_chars = self.options.text
        if isinstance(unicode_chars, bytes):
            unicode_chars = unicode_chars.decode(self.encoding)

        # TODO: remove duplicate chars

        for char in unicode_chars:
            # Create a new layer.
            layer = self.svg.add(inkex.Layer.new("GlyphLayer-" + char))
            layer.set("style", "display:none")  # initially not visible

            # TODO: make it optional ("Use current selection as template glyph")
            # Move selection to the newly created layer
            for node in self.svg.selection.values():
                layer.append(node)


if __name__ == "__main__":
    NewGlyphLayer().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2005,2007 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=ungrouped-imports
"""
Embed images so they are base64 encoded data inside the svg.
"""

from __future__ import unicode_literals

import os

import inkex
from inkex import Image
from inkex.localization import inkex_gettext as _

try:
    import urllib.request as urllib
    import urllib.parse as urlparse
    from base64 import encodebytes
except ImportError:
    # python2 compatibility, remove when python3 only.
    import urllib
    import urlparse
    from base64 import encodestring as encodebytes


class EmbedImage(inkex.EffectExtension):
    """Allow selected image tags to become embeded image tags"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--selectedonly", type=inkex.Boolean, help="embed only selected images"
        )

    def effect(self):
        # if slectedonly is enabled and there is a selection
        # only embed selected images. otherwise embed all images
        if self.options.selectedonly:
            images = self.svg.selection.get(Image)
        else:
            images = self.svg.xpath("//svg:image")

        for node in images:
            self.embed_image(node)

    def embed_image(self, node):
        """Embed the data of the selected Image Tag element"""
        xlink = node.get("xlink:href")
        if xlink is not None and xlink[:5] == "data:":
            # No need, data already embedded
            return
        if xlink is None:
            inkex.errormsg(
                _('Attribute "xlink:href" not set on node {}.'.format(node.get_id()))
            )
            return

        url = urlparse.urlparse(xlink)
        href = urllib.url2pathname(url.path)

        # Look relative to the *temporary* filename instead of the original filename.
        path = self.absolute_href(
            href or "", cwd=os.path.dirname(self.options.input_file)
        )

        # Backup directory where we can find the image
        if not os.path.isfile(path):
            path = node.get("sodipodi:absref", path)

        if not os.path.isfile(path):
            inkex.errormsg(
                _('File not found "{}". Unable to embed image.').format(path)
            )
            return

        with open(path, "rb") as handle:
            # Don't read the whole file to check the header
            file_type = get_type(path, handle.read(10))
            handle.seek(0)

            if file_type:
                # Future: Change encodestring to encodebytes when python3 only
                node.set(
                    "xlink:href",
                    "data:{};base64,{}".format(
                        file_type, encodebytes(handle.read()).decode("ascii")
                    ),
                )
                node.pop("sodipodi:absref")
            else:
                inkex.errormsg(
                    _(
                        "%s is not of type image/png, image/jpeg, "
                        "image/bmp, image/gif, image/tiff, or image/x-icon"
                    )
                    % path
                )


def get_type(path, header):
    """Basic magic header checker, returns mime type"""
    for head, mime in (
        (b"\x89PNG", "image/png"),
        (b"\xff\xd8", "image/jpeg"),
        (b"BM", "image/bmp"),
        (b"GIF87a", "image/gif"),
        (b"GIF89a", "image/gif"),
        (b"MM\x00\x2a", "image/tiff"),
        (b"II\x2a\x00", "image/tiff"),
    ):
        if header.startswith(head):
            return mime

    # ico files lack any magic... therefore we check the filename instead
    for ext, mime in (
        # official IANA registered MIME is 'image/vnd.microsoft.icon' tho
        (".ico", "image/x-icon"),
        (".svg", "image/svg+xml"),
    ):
        if path.endswith(ext):
            return mime
    return None


if __name__ == "__main__":
    EmbedImage().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
When the extension manager fails, this allows us to upgrade automatically.
"""

import os
import sys
import traceback

# Note: All imports are being done as late as possible in this module.
# If any of the imports caused the error, then we'll still message the user.


def update_pip_package(package):
    """
    Update the package using the pip update.
    """
    from inkex.utils import get_user_directory
    from inkex.command import ProgramRunError, call

    try:
        pip = os.path.join(get_user_directory(), "bin", "pip")
        log = call(pip, "install", "--upgrade", package).decode("utf8")
        logs = [line for line in log.split("\n") if "skip" not in line]
        return "\n".join(logs)
    except ProgramRunError as err:
        raise IOError(f"Failed to update the package {package}")


def attempt_to_recover():
    """
    Messages the user, provides a traceback and attepts a self-update
    """
    info = sys.exc_info()
    sys.stderr.write("An error occured with the extensions manager!\n")
    sys.stderr.write("Trying to self-update the package... ")
    sys.stderr.flush()

    from inkman import __pkgname__, __version__, __file__, __issues__

    location = os.path.dirname(__file__)
    update_log = "Not done"

    try:
        update_log = update_pip_package(__pkgname__)
        sys.stderr.write("Updated!\n\nPlease try and reload the program again.\n\n")
    except Exception:
        sys.stderr.write(
            "Failed to update!\n\nPlease delete the package manually! (see location below)\n\n"
        )

    sys.stderr.write(
        f"""
Please report this error
------------------------

Report URL: {__issues__}
Location: {location}
{__pkgname__}: {__version__}

{update_log}
"""
    )

    traceback.print_exception(*info)
    del info
    sys.exit(2)



#!/usr/bin/env python
#
# Copyright (C) 2010 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from lxml import etree

import inkex
from webslicer_effect import WebSlicerMixin, is_empty


class CreateRect(WebSlicerMixin, inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--name")
        pars.add_argument("--format", default="png")
        pars.add_argument("--dpi", type=float)
        pars.add_argument("--dimension")
        pars.add_argument("--bg-color")
        pars.add_argument("--quality", type=int)
        pars.add_argument("--gif-type")
        pars.add_argument("--palette-size", type=int)
        pars.add_argument("--html-id")
        pars.add_argument("--html-class")
        pars.add_argument("--layout-disposition")
        pars.add_argument("--layout-position-anchor")
        # inkscape param workaround
        pars.add_argument("--tab")

    def unique_slice_name(self):
        name = self.options.name
        el = self.svg.xpath('//*[@id="' + name + '"]')
        if len(el) > 0:
            if name[-3:] == "-00":
                name = name[:-3]
            num = 0
            num_s = "00"
            while len(el) > 0:
                num += 1
                num_s = str(num)
                if len(num_s) == 1:
                    num_s = "0" + num_s
                el = self.svg.xpath('//*[@id="' + name + "-" + num_s + '"]')
            self.options.name = name + "-" + num_s

    def validate_options(self):
        self.options.format = self.options.format.lower()
        if not is_empty(self.options.dimension):
            self.options.dimension

    def effect(self):
        scale = self.svg.unittouu("1px")  # convert to document units
        self.validate_options()
        layer = self.get_slicer_layer(True)
        # TODO: get selected elements to define location and size
        rect = etree.SubElement(layer, "rect")
        if is_empty(self.options.name):
            self.options.name = "slice-00"
        self.unique_slice_name()
        rect.set("id", self.options.name)
        rect.set("fill", "red")
        rect.set("opacity", "0.5")
        rect.set("x", str(-scale * 100))
        rect.set("y", str(-scale * 100))
        rect.set("width", str(scale * 200))
        rect.set("height", str(scale * 200))
        desc = etree.SubElement(rect, "desc")
        conf_txt = "format:" + self.options.format + "\n"
        if not is_empty(self.options.dpi):
            conf_txt += "dpi:" + str(self.options.dpi) + "\n"
        if not is_empty(self.options.html_id):
            conf_txt += "html-id:" + self.options.html_id
        desc.text = self.get_conf_text_from_list(self.get_conf_list())

    def get_conf_list(self):
        conf_list = ["format"]
        if self.options.format == "gif":
            conf_list.extend(["gif_type", "palette_size"])
        if self.options.format == "jpg":
            conf_list.extend(["quality"])
        conf_list.extend(
            [
                "dpi",
                "dimension",
                "bg_color",
                "html_id",
                "html_class",
                "layout_disposition",
                "layout_position_anchor",
            ]
        )
        return conf_list


if __name__ == "__main__":
    CreateRect().run()



#! /usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 Joel Holdsworth joel@airwebreathe.org.uk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import math
import inkex


class Spirograph(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument(
            "--primaryr", type=float, default=100.0, help="The radius of the outer gear"
        )
        pars.add_argument(
            "--secondaryr",
            type=float,
            default=60.0,
            help="The radius of the inner gear",
        )
        pars.add_argument(
            "--penr",
            type=float,
            default=50.0,
            help="The distance of the pen from the inner gear",
        )
        pars.add_argument(
            "--gearplacement",
            default="inside",
            help="Selects whether the gear is inside or outside the ring",
        )
        pars.add_argument(
            "--rotation",
            type=float,
            default=0.0,
            help="The number of degrees to rotate the image by",
        )
        pars.add_argument(
            "--quality",
            type=int,
            default=16,
            help="The quality of the calculated output",
        )

    def effect(self):
        self.options.primaryr = self.svg.unittouu(str(self.options.primaryr) + "px")
        self.options.secondaryr = self.svg.unittouu(str(self.options.secondaryr) + "px")
        self.options.penr = self.svg.unittouu(str(self.options.penr) + "px")

        if self.options.secondaryr == 0:
            return
        if self.options.quality == 0:
            return

        if self.options.gearplacement.strip(" ").lower().startswith("outside"):
            a = self.options.primaryr + self.options.secondaryr
            flip = -1
        else:
            a = self.options.primaryr - self.options.secondaryr
            flip = 1

        ratio = a / self.options.secondaryr
        if ratio == 0:
            return
        scale = 2 * math.pi / (ratio * self.options.quality)

        rotation = -math.pi * self.options.rotation / 180

        new = inkex.PathElement()
        new.style = inkex.Style(stroke="#000000", fill="none", stroke_width="1.0")

        path_string = ""
        maxPointCount = 1000

        for i in range(maxPointCount):

            theta = i * scale

            view_center = self.svg.namedview.center
            x = (
                a * math.cos(theta + rotation)
                + self.options.penr * math.cos(ratio * theta + rotation) * flip
                + view_center[0]
            )
            y = (
                a * math.sin(theta + rotation)
                - self.options.penr * math.sin(ratio * theta + rotation)
                + view_center[1]
            )

            dx = (
                (
                    -a * math.sin(theta + rotation)
                    - ratio
                    * self.options.penr
                    * math.sin(ratio * theta + rotation)
                    * flip
                )
                * scale
                / 3
            )
            dy = (
                (
                    a * math.cos(theta + rotation)
                    - ratio * self.options.penr * math.cos(ratio * theta + rotation)
                )
                * scale
                / 3
            )

            if i <= 0:
                path_string += "M {},{} C {},{} ".format(
                    str(x), str(y), str(x + dx), str(y + dy)
                )
            else:
                path_string += "{},{} {},{}".format(
                    str(x - dx), str(y - dy), str(x), str(y)
                )

                if (
                    math.fmod(i / ratio, self.options.quality) == 0
                    and i % self.options.quality == 0
                ):
                    path_string += "Z"
                    break
                else:
                    if i == maxPointCount - 1:
                        pass  # we reached the allowed maximum of points, stop here
                    else:
                        path_string += " C {},{} ".format(str(x + dx), str(y + dy))

        new.path = path_string
        self.svg.get_current_layer().append(new)


if __name__ == "__main__":
    Spirograph().run()



#
# Copyright 2018-2022 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""All graphical interfaces bound into a GtkApp"""

import os

from inkex.gui import GtkApp

from .main import ExtensionManagerWindow
from .info import MoreInformation


class ManagerApp(GtkApp):
    """Load the inkscape extensions glade file and attach to window"""

    ui_dir = os.path.join(os.path.dirname(__file__), "..", "data")
    app_name = "inkscape-extensions-manager"
    windows = [ExtensionManagerWindow, MoreInformation]



#
# Copyright 2011 (c) Ian Bicking <ianb@colorstudy.com>
#           2019 (c) Martin Owens <doctormo@gmail.com>
#
# Taken from http://formencode.org under the GPL compatible PSF License.
# Modified to produce more output as a diff.
#
"""
Allow two xml files/lxml etrees to be compared, returning their differences.
"""
import xml.etree.ElementTree as xml
from io import BytesIO

from inkex.paths import Path


def text_compare(test1, test2):
    """
    Compare two text strings while allowing for '*' to match
    anything on either lhs or rhs.
    """
    if not test1 and not test2:
        return True
    if test1 == "*" or test2 == "*":
        return True
    return (test1 or "").strip() == (test2 or "").strip()


class DeltaLogger(list):
    """A record keeper of the delta between two svg files"""

    def append_tag(self, tag_a, tag_b):
        """Record a tag difference"""
        if tag_a:
            tag_a = f"<{tag_a}.../>"
        if tag_b:
            tag_b = f"<{tag_b}.../>"
        self.append((tag_a, tag_b))

    def append_attr(self, attr, value_a, value_b):
        """Record an attribute difference"""

        def _prep(val):
            if val:
                if attr == "d":
                    return [attr] + Path(val).to_arrays()
                return (attr, val)
            return val

        # Only append a difference if the preprocessed values are different.
        # This solves the issue that -0 != 0 in path data.
        prep_a = _prep(value_a)
        prep_b = _prep(value_b)
        if prep_a != prep_b:
            self.append((prep_a, prep_b))

    def append_text(self, text_a, text_b):
        """Record a text difference"""
        self.append((text_a, text_b))

    def __bool__(self):
        """Returns True if there's no log, i.e. the delta is clean"""
        return not self.__len__()

    __nonzero__ = __bool__

    def __repr__(self):
        if self:
            return "No differences detected"
        return f"{len(self)} xml differences"


def to_xml(data):
    """Convert string or bytes to xml parsed root node"""
    if isinstance(data, str):
        data = data.encode("utf8")
    if isinstance(data, bytes):
        return xml.parse(BytesIO(data)).getroot()
    return data


def xmldiff(data1, data2):
    """Create an xml difference, will modify the first xml structure with a diff"""
    xml1, xml2 = to_xml(data1), to_xml(data2)
    delta = DeltaLogger()
    _xmldiff(xml1, xml2, delta)
    return xml.tostring(xml1).decode("utf-8"), delta


def _xmldiff(xml1, xml2, delta):
    if xml1.tag != xml2.tag:
        xml1.tag = f"{xml1.tag}XXX{xml2.tag}"
        delta.append_tag(xml1.tag, xml2.tag)
    for name, value in xml1.attrib.items():
        if name not in xml2.attrib:
            delta.append_attr(name, xml1.attrib[name], None)
            xml1.attrib[name] += "XXX"
        elif xml2.attrib.get(name) != value:
            delta.append_attr(name, xml1.attrib.get(name), xml2.attrib.get(name))
            xml1.attrib[name] = f"{xml1.attrib.get(name)}XXX{xml2.attrib.get(name)}"
    for name, value in xml2.attrib.items():
        if name not in xml1.attrib:
            delta.append_attr(name, None, value)
            xml1.attrib[name] = "XXX" + value
    if not text_compare(xml1.text, xml2.text):
        delta.append_text(xml1.text, xml2.text)
        xml1.text = f"{xml1.text}XXX{xml2.text}"
    if not text_compare(xml1.tail, xml2.tail):
        delta.append_text(xml1.tail, xml2.tail)
        xml1.tail = f"{xml1.tail}XXX{xml2.tail}"

    # Get children and pad with nulls
    children_a = list(xml1)
    children_b = list(xml2)
    children_a += [None] * (len(children_b) - len(children_a))
    children_b += [None] * (len(children_a) - len(children_b))

    for child_a, child_b in zip(children_a, children_b):
        if child_a is None:  # child_b exists
            delta.append_tag(child_b.tag, None)
        elif child_b is None:  # child_a exists
            delta.append_tag(None, child_a.tag)
        else:
            _xmldiff(child_a, child_b, delta)
